# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# classifiers.py
# Copyright (C) 2014-2016 Fracpete (pythonwekawrapper at gmail dot com)

import unittest
import weka.core.jvm as jvm
from weka.core.classes import Random
import weka.core.converters as converters
import weka.classifiers as classifiers
import weka.filters as filters
import weka.plot.classifiers as plot
import wekatests.tests.weka_test as weka_test


class TestClassifiers(weka_test.WekaTest):

    def test_generate_thresholdcurve_data(self):
        """
        Tests the generate_thresholdcurve_data method.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("diabetes.arff"))
        data.class_is_last()

        remove = filters.Filter(classname="weka.filters.unsupervised.attribute.Remove", options=["-R", "1-3"])
        cls = classifiers.Classifier(classname="weka.classifiers.bayes.NaiveBayes")
        fc = classifiers.FilteredClassifier()
        fc.filter = remove
        fc.classifier = cls

        evl = classifiers.Evaluation(data)
        evl.crossvalidate_model(cls, data, 10, Random(1))
        data = plot.generate_thresholdcurve_data(evl, 0)
        self.assertEqual(13, data.num_attributes, msg="number of attributes differs")
        self.assertEqual(769, data.num_instances, msg="number of rows differs")
        attname = "True Positives"
        self.assertIsNotNone(data.attribute_by_name(attname), msg="Failed to locate attribute: " + attname)
        attname = "False Positive Rate"
        self.assertIsNotNone(data.attribute_by_name(attname), msg="Failed to locate attribute: " + attname)
        attname = "Lift"
        self.assertIsNotNone(data.attribute_by_name(attname), msg="Failed to locate attribute: " + attname)

    def test_get_auc(self):
        """
        Tests the get_auc method.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("diabetes.arff"))
        data.class_is_last()

        remove = filters.Filter(classname="weka.filters.unsupervised.attribute.Remove", options=["-R", "1-3"])
        cls = classifiers.Classifier(classname="weka.classifiers.bayes.NaiveBayes")
        fc = classifiers.FilteredClassifier()
        fc.filter = remove
        fc.classifier = cls

        evl = classifiers.Evaluation(data)
        evl.crossvalidate_model(cls, data, 10, Random(1))
        data = plot.generate_thresholdcurve_data(evl, 0)
        area = plot.get_auc(data)
        self.assertAlmostEqual(0.819, area, places=3, msg="AUC differs")

    def test_get_prc(self):
        """
        Tests the get_prc method.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("diabetes.arff"))
        data.class_is_last()

        remove = filters.Filter(classname="weka.filters.unsupervised.attribute.Remove", options=["-R", "1-3"])
        cls = classifiers.Classifier(classname="weka.classifiers.bayes.NaiveBayes")
        fc = classifiers.FilteredClassifier()
        fc.filter = remove
        fc.classifier = cls

        evl = classifiers.Evaluation(data)
        evl.crossvalidate_model(cls, data, 10, Random(1))
        data = plot.generate_thresholdcurve_data(evl, 0)
        area = plot.get_prc(data)
        self.assertAlmostEqual(0.892, area, places=3, msg="PRC differs")

    def test_plot_classifier_errors(self):
        """
        Tests the plot_classifier_errors method.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        bolts_data = loader.load_file(self.datafile("bolts.arff"))
        self.assertIsNotNone(bolts_data)
        bolts_data.class_is_last()

        classifier = classifiers.Classifier(classname="weka.classifiers.functions.LinearRegression", options=["-S", "1", "-C"])
        evaluation = classifiers.Evaluation(bolts_data)
        evaluation.crossvalidate_model(classifier, bolts_data, 10, Random(42))
        plot.plot_classifier_errors(evaluation.predictions, wait=False)

    def test_plot_roc(self):
        """
        Tests the plot_roc method.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("diabetes.arff"))
        data.class_is_last()

        remove = filters.Filter(classname="weka.filters.unsupervised.attribute.Remove", options=["-R", "1-3"])
        cls = classifiers.Classifier(classname="weka.classifiers.bayes.NaiveBayes")
        fc = classifiers.FilteredClassifier()
        fc.filter = remove
        fc.classifier = cls

        evl = classifiers.Evaluation(data)
        evl.crossvalidate_model(cls, data, 10, Random(1))
        plot.plot_roc(evl, class_index=[0, 1], wait=False)

    def test_plot_prc(self):
        """
        Tests the plot_prc method.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("diabetes.arff"))
        data.class_is_last()

        remove = filters.Filter(classname="weka.filters.unsupervised.attribute.Remove", options=["-R", "1-3"])
        cls = classifiers.Classifier(classname="weka.classifiers.bayes.NaiveBayes")
        fc = classifiers.FilteredClassifier()
        fc.filter = remove
        fc.classifier = cls

        evl = classifiers.Evaluation(data)
        evl.crossvalidate_model(cls, data, 10, Random(1))
        plot.plot_prc(evl, class_index=[0, 1], wait=False)

    def test_learning_curve(self):
        """
        Tests the learning_curve method.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("diabetes.arff"))
        data.class_is_last()

        cls = [
            classifiers.Classifier(classname="weka.classifiers.trees.J48"),
            classifiers.Classifier(classname="weka.classifiers.bayes.NaiveBayesUpdateable")]
        plot.plot_learning_curve(cls, data, wait=False)
        plot.plot_learning_curve(cls, data, increments=0.1, wait=False)


def suite():
    """
    Returns the test suite.
    :return: the test suite
    :rtype: unittest.TestSuite
    """
    return unittest.TestLoader().loadTestsFromTestCase(TestClassifiers)


if __name__ == '__main__':
    jvm.start()
    unittest.TextTestRunner().run(suite())
    jvm.stop()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# classifiers.py
# Copyright (C) 2014-2021 Fracpete (pythonwekawrapper at gmail dot com)

import sys
import os
import javabridge
import logging
import argparse
import traceback
import weka.core.jvm as jvm
import weka.core.typeconv as typeconv
import weka.core.classes as classes
import weka.core.serialization as serialization
from numpy import *
from weka.core.classes import JavaObject, join_options, OptionHandler, Random, SelectedTag, Tags, Tag, JavaArray, \
    is_instance_of
from weka.core.classes import AbstractParameter
from weka.core.capabilities import Capabilities
from weka.core.dataset import Instances, Instance, Attribute
from weka.filters import Filter
from weka.attribute_selection import ASEvaluation, ASSearch

# logging setup
logger = logging.getLogger("weka.classifiers")


class Classifier(OptionHandler):
    """
    Wrapper class for classifiers.
    """

    def __init__(self, classname="weka.classifiers.rules.ZeroR", jobject=None, options=None):
        """
        Initializes the specified classifier using either the classname or the supplied JB_Object.

        :param classname: the classname of the classifier
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        if jobject is None:
            jobject = Classifier.new_instance(classname)
        self.enforce_type(jobject, "weka.classifiers.Classifier")
        self.is_updateable = self.check_type(jobject, "weka.classifiers.UpdateableClassifier")
        self.is_drawable = self.check_type(jobject, "weka.core.Drawable")
        self.is_batchpredictor = self.check_type(jobject, "weka.core.BatchPredictor")
        super(Classifier, self).__init__(jobject=jobject, options=options)
        self.__classify = javabridge.make_call(self.jobject, "classifyInstance", "(Lweka/core/Instance;)D")
        self.__distribution = javabridge.make_call(self.jobject, "distributionForInstance", "(Lweka/core/Instance;)[D")
        if self.is_batchpredictor:
            self.__distributions = javabridge.make_call(
                self.jobject, "distributionsForInstances", "(Lweka/core/Instances;)[[D")

    @property
    def capabilities(self):
        """
        Returns the capabilities of the classifier.

        :return: the capabilities
        :rtype: Capabilities
        """
        return Capabilities(javabridge.call(self.jobject, "getCapabilities", "()Lweka/core/Capabilities;"))

    def build_classifier(self, data):
        """
        Builds the classifier with the data.

        :param data: the data to train the classifier with
        :type data: Instances
        """
        javabridge.call(self.jobject, "buildClassifier", "(Lweka/core/Instances;)V", data.jobject)

    def update_classifier(self, inst):
        """
        Updates the classifier with the instance.

        :param inst: the Instance to update the classifier with
        :type inst: Instance
        """
        if self.is_updateable:
            javabridge.call(self.jobject, "updateClassifier", "(Lweka/core/Instance;)V", inst.jobject)
        else:
            logger.critical(classes.get_classname(self.jobject) + " is not updateable!")

    def classify_instance(self, inst):
        """
        Peforms a prediction.

        :param inst: the Instance to get a prediction for
        :type inst: Instance
        :return: the classification (either regression value or 0-based label index)
        :rtype: float
        """
        return self.__classify(inst.jobject)

    def distribution_for_instance(self, inst):
        """
        Peforms a prediction, returning the class distribution.

        :param inst: the Instance to get the class distribution for
        :type inst: Instance
        :return: the class distribution array
        :rtype: ndarray
        """
        pred = self.__distribution(inst.jobject)
        return javabridge.get_env().get_double_array_elements(pred)

    def distributions_for_instances(self, data):
        """
        Peforms predictions, returning the class distributions.

        :param data: the Instances to get the class distributions for
        :type data: Instances
        :return: the class distribution matrix, None if not a batch predictor
        :rtype: ndarray
        """
        if self.is_batchpredictor:
            return typeconv.jdouble_matrix_to_ndarray(self.__distributions(data.jobject))
        else:
            return None

    @property
    def batch_size(self):
        """
        Returns the batch size, in case this classifier is a batch predictor.

        :return: the batch size, None if not a batch predictor
        :rtype: str
        """
        if self.is_batchpredictor:
            return javabridge.call(self.jobject, "getBatchSize", "()Ljava/lang/String;")
        else:
            return None

    @batch_size.setter
    def batch_size(self, size):
        """
        Sets the batch size, in case this classifier is a batch predictor.

        :param size: the size of the batch
        :type size: str
        """
        if self.is_batchpredictor:
            javabridge.call(self.jobject, "setBatchSize", "(Ljava/lang/String;)V", size)

    def has_efficient_batch_prediction(self):
        """
        Returns whether the classifier implements a more efficient batch prediction.

        :return: True if a more efficient batch prediction is implemented, always False if not batch predictor
        :rtype: bool
        """
        if self.is_batchpredictor:
            return javabridge.call(self.jobject, "implementsMoreEfficientBatchPrediction", "()Z")
        else:
            return False

    @property
    def graph_type(self):
        """
        Returns the graph type if classifier implements weka.core.Drawable, otherwise -1.

        :return: the type
        :rtype: int
        """
        if self.is_drawable:
            return javabridge.call(self.jobject, "graphType", "()I")
        else:
            return -1

    @property
    def graph(self):
        """
        Returns the graph if classifier implements weka.core.Drawable, otherwise None.

        :return: the generated graph string
        :rtype: str
        """
        if self.is_drawable:
            return javabridge.call(self.jobject, "graph", "()Ljava/lang/String;")
        else:
            return None

    def to_source(self, classname):
        """
        Returns the model as Java source code if the classifier implements weka.classifiers.Sourcable.

        :param classname: the classname for the generated Java code
        :type classname: str
        :return: the model as source code string
        :rtype: str
        """
        if not self.check_type(self.jobject, "weka.classifiers.Sourcable"):
            return None
        return javabridge.call(self.jobject, "toSource", "(Ljava/lang/String;)Ljava/lang/String;", classname)

    @classmethod
    def make_copy(cls, classifier):
        """
        Creates a copy of the classifier.

        :param classifier: the classifier to copy
        :type classifier: Classifier
        :return: the copy of the classifier
        :rtype: Classifier
        """
        return Classifier(
            jobject=javabridge.static_call(
                "weka/classifiers/AbstractClassifier", "makeCopy",
                "(Lweka/classifiers/Classifier;)Lweka/classifiers/Classifier;", classifier.jobject))

    @classmethod
    def deserialize(cls, ser_file):
        """
        Deserializes a classifier from a file.

        :param ser_file: the model file to deserialize
        :type ser_file: str
        :return: model and, if available, the dataset header
        :rtype: tuple
        """

        objs = serialization.read_all(ser_file)
        if len(objs) == 1:
            return Classifier(jobject=objs[0]), None
        elif len(objs) == 2:
            return Classifier(jobject=objs[0]), Instances(jobject=objs[1])
        else:
            raise Exception(
                "Excepted one or two objects in the model file (%s), but encountered: %d" % (ser_file, len(objs)))

    def serialize(self, ser_file, header=None):
        """
        Serializes the classifier to the specified file.

        :param ser_file: the file to save the model to
        :type ser_file: str
        :param header: the (optional) dataset header to store alongside; recommended
        :type header: Instances
        """

        if (header is not None) and header.num_instances > 0:
            header = Instances.template_instances(header)

        if header is not None:
            serialization.write_all(ser_file, [self, header])
        else:
            serialization.write(ser_file, self)


class SingleClassifierEnhancer(Classifier):
    """
    Wrapper class for classifiers that use a single base classifier.
    """

    def __init__(self, classname=None, jobject=None, options=None):
        """
        Initializes the specified classifier using either the classname or the supplied JB_Object.

        :param classname: the classname of the classifier
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        if jobject is None:
            jobject = Classifier.new_instance(classname)
        self.enforce_type(jobject, "weka.classifiers.SingleClassifierEnhancer")
        super(SingleClassifierEnhancer, self).__init__(classname=classname, jobject=jobject, options=options)

    @property
    def classifier(self):
        """
        Returns the base classifier.

        ;return: the base classifier
        :rtype: Classifier
        """
        return Classifier(jobject=javabridge.call(self.jobject, "getClassifier", "()Lweka/classifiers/Classifier;"))

    @classifier.setter
    def classifier(self, classifier):
        """
        Sets the base classifier.

        :param classifier: the base classifier to use
        :type classifier: Classifier
        """
        javabridge.call(self.jobject, "setClassifier", "(Lweka/classifiers/Classifier;)V", classifier.jobject)


class FilteredClassifier(SingleClassifierEnhancer):
    """
    Wrapper class for the filtered classifier.
    """

    def __init__(self, jobject=None, options=None):
        """
        Initializes the specified classifier using its classname or the supplied JB_Object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        classname = "weka.classifiers.meta.FilteredClassifier"
        if jobject is None:
            jobject = Classifier.new_instance(classname)
        else:
            self.enforce_type(jobject, classname)
        super(FilteredClassifier, self).__init__(jobject=jobject, options=options)

    @property
    def filter(self):
        """
        Returns the filter.

        :return: the filter in use
        :rtype: Filter
        """
        return Filter(jobject=javabridge.call(self.jobject, "getFilter", "()Lweka/filters/Filter;"))

    @filter.setter
    def filter(self, filtr):
        """
        Sets the filter.

        :param filtr: the filter to use
        :type filtr: Filter
        """
        javabridge.call(self.jobject, "setFilter", "(Lweka/filters/Filter;)V", filtr.jobject)

    def check_for_modified_class_attribute(self, check):
        """
        Sets whether to check for class attribute modifications.

        :param check: True if checking for modifications
        :type check: bool
        """
        javabridge.call(self.jobject, "setDoNotCheckForModifiedClassAttribute", "(Z)V", not check)


class AttributeSelectedClassifier(SingleClassifierEnhancer):
    """
    Wrapper class for the AttributeSelectedClassifier.
    """

    def __init__(self, jobject=None, options=None):
        """
        Initializes the specified classifier using its classname or the supplied JB_Object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        classname = "weka.classifiers.meta.AttributeSelectedClassifier"
        if jobject is None:
            jobject = Classifier.new_instance(classname)
        else:
            self.enforce_type(jobject, classname)
        super(AttributeSelectedClassifier, self).__init__(jobject=jobject, options=options)

    @property
    def evaluator(self):
        """
        Returns the evaluator.

        :return: the evaluator in use
        :rtype: ASEvaluation
        """
        return ASEvaluation(jobject=javabridge.call(self.jobject, "getEvaluator", "()Lweka/attributeSelection/ASEvaluation;"))

    @evaluator.setter
    def evaluator(self, evl):
        """
        Sets the evaluator.

        :param evl: the evaluator to use
        :type evl: ASEvaluation
        """
        javabridge.call(self.jobject, "setEvaluator", "(Lweka/attributeSelection/ASEvaluation;)V", evl.jobject)

    @property
    def search(self):
        """
        Returns the search.

        :return: the search in use
        :rtype: ASSearch
        """
        return ASSearch(jobject=javabridge.call(self.jobject, "getSearch", "()Lweka/attributeSelection/ASSearch;"))

    @search.setter
    def search(self, search):
        """
        Sets the search.

        :param search: the search to use
        :type search: ASSearch
        """
        javabridge.call(self.jobject, "setSearch", "(Lweka/attributeSelection/ASSearch;)V", search.jobject)


class GridSearch(SingleClassifierEnhancer):
    """
    Wrapper class for the GridSearch meta-classifier.
    """

    def __init__(self, jobject=None, options=None):
        """
        Initializes the specified classifier using its classname or the supplied JB_Object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        classname = "weka.classifiers.meta.GridSearch"
        if jobject is None:
            jobject = GridSearch.new_instance(classname)
            if jobject is None:
                raise Exception(
                    "Failed to instantiate GridSearch - package installed and jvm started with package support?")
        else:
            self.enforce_type(jobject, classname)
        super(GridSearch, self).__init__(jobject=jobject, options=options)
        self.tags_evaluation = Tags.get_tags("weka.classifiers.meta.GridSearch", "TAGS_EVALUATION")

    @property
    def evaluation(self):
        """
        Returns the currently set statistic used for evaluation.

        :return: the statistic
        :rtype: SelectedTag
        """
        return SelectedTag(
            javabridge.call(self.jobject, "getEvaluation", "()Lweka/core/SelectedTag;"))

    @evaluation.setter
    def evaluation(self, evl):
        """
        Sets the statistic to use for evaluation.

        :param evl: the statistic
        :type evl: SelectedTag, Tag or str
        """
        if isinstance(evl, str):
            evl = self.tags_evaluation.find(evl)
        if isinstance(evl, Tag):
            evl = SelectedTag(tag_id=evl.ident, tags=self.tags_evaluation)
        javabridge.call(self.jobject, "setEvaluation", "(Lweka/core/SelectedTag;)V", evl.jobject)

    @property
    def x(self):
        """
        Returns a dictionary with all the current values for the X of the grid.
        Keys for the dictionary: property, min, max, step, base, expression
        Types: property=str, min=float, max=float, step=float, base=float, expression=str

        :return: the dictionary with the parameters
        :rtype: dict
        """
        result = {}
        result["property"] = javabridge.call(self.jobject, "getXProperty", "()Ljava/lang/String;")
        result["min"] = javabridge.call(self.jobject, "getXMin", "()D")
        result["max"] = javabridge.call(self.jobject, "getXMax", "()D")
        result["step"] = javabridge.call(self.jobject, "getXStep", "()D")
        result["base"] = javabridge.call(self.jobject, "getXBase", "()D")
        result["expression"] = javabridge.call(self.jobject, "getXExpression", "()Ljava/lang/String;")
        return result

    @x.setter
    def x(self, d):
        """
        Allows to configure the X of the grid with one method call.
        Keys for the dictionary: property, min, max, step, base, expression
        Types: property=str, min=float, max=float, step=float, base=float, expression=str

        :param d: the dictionary with the parameters
        :type d: dict
        """
        if "property" in d:
            javabridge.call(self.jobject, "setXProperty", "(Ljava/lang/String;)V", d["property"])
        if "min" in d:
            javabridge.call(self.jobject, "setXMin", "(D)V", d["min"])
        if "max" in d:
            javabridge.call(self.jobject, "setXMax", "(D)V", d["max"])
        if "step" in d:
            javabridge.call(self.jobject, "setXStep", "(D)V", d["step"])
        if "base" in d:
            javabridge.call(self.jobject, "setXBase", "(D)V", d["base"])
        if "expression" in d:
            javabridge.call(self.jobject, "setXExpression", "(Ljava/lang/String;)V", d["expression"])

    @property
    def y(self):
        """
        Returns a dictionary with all the current values for the Y of the grid.
        Keys for the dictionary: property, min, max, step, base, expression
        Types: property=str, min=float, max=float, step=float, base=float, expression=str

        :return: the dictionary with the parameters
        :rtype: dict
        """
        result = {}
        result["property"] = javabridge.call(self.jobject, "getYProperty", "()Ljava/lang/String;")
        result["min"] = javabridge.call(self.jobject, "getYMin", "()D")
        result["max"] = javabridge.call(self.jobject, "getYMax", "()D")
        result["step"] = javabridge.call(self.jobject, "getYStep", "()D")
        result["base"] = javabridge.call(self.jobject, "getYBase", "()D")
        result["expression"] = javabridge.call(self.jobject, "getYExpression", "()Ljava/lang/String;")
        return result

    @y.setter
    def y(self, d):
        """
        Allows to configure the Y of the grid with one method call.
        Keys for the dictionary: property, min, max, step, base, expression
        Types: property=str, min=float, max=float, step=float, base=float, expression=str

        :param d: the dictionary with the parameters
        :type d: dict
        """
        if "property" in d:
            javabridge.call(self.jobject, "setYProperty", "(Ljava/lang/String;)V", d["property"])
        if "min" in d:
            javabridge.call(self.jobject, "setYMin", "(D)V", d["min"])
        if "max" in d:
            javabridge.call(self.jobject, "setYMax", "(D)V", d["max"])
        if "step" in d:
            javabridge.call(self.jobject, "setYStep", "(D)V", d["step"])
        if "base" in d:
            javabridge.call(self.jobject, "setYBase", "(D)V", d["base"])
        if "expression" in d:
            javabridge.call(self.jobject, "setYExpression", "(Ljava/lang/String;)V", d["expression"])

    @property
    def best(self):
        """
        Returns the best classifier setup found during the th search.

        :return: the best classifier setup
        :rtype: Classifier
        """
        return Classifier(jobject=javabridge.call(self.jobject, "getBestClassifier", "()Lweka/classifiers/Classifier;"))


class MultiSearch(SingleClassifierEnhancer):
    """
    Wrapper class for the MultiSearch meta-classifier.
    NB: 'multi-search-weka-package' must be installed (https://github.com/fracpete/multisearch-weka-package),
    version 2016.1.15 or later.
    """

    def __init__(self, jobject=None, options=None):
        """
        Initializes the specified classifier using its classname or the supplied JB_Object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        classname = "weka.classifiers.meta.MultiSearch"
        if jobject is None:
            jobject = MultiSearch.new_instance(classname)
        else:
            self.enforce_type(jobject, classname)
        super(MultiSearch, self).__init__(jobject=jobject, options=options)
        self.tags_evaluation = Tags.get_object_tags(self, "getMetricsTags")

    @property
    def evaluation(self):
        """
        Returns the currently set statistic used for evaluation.

        :return: the statistic
        :rtype: SelectedTag
        """
        return SelectedTag(
            javabridge.call(self.jobject, "getEvaluation", "()Lweka/core/SelectedTag;"))

    @evaluation.setter
    def evaluation(self, evl):
        """
        Sets the statistic to use for evaluation.

        :param evl: the statistic
        :type evl: SelectedTag, Tag or str
        """
        if isinstance(evl, str):
            evl = self.tags_evaluation.find(evl)
        if isinstance(evl, Tag):
            evl = SelectedTag(tag_id=evl.ident, tags=self.tags_evaluation)
        javabridge.call(self.jobject, "setEvaluation", "(Lweka/core/SelectedTag;)V", evl.jobject)

    @property
    def parameters(self):
        """
        Returns the list of currently set search parameters.

        :return: the list of AbstractSearchParameter objects
        :rtype: list
        """
        array = JavaArray(
            javabridge.call(self.jobject, "getSearchParameters", "()[Lweka/core/setupgenerator/AbstractParameter;"))
        result = []
        for item in array:
            result.append(AbstractParameter(jobject=item.jobject))
        return result

    @parameters.setter
    def parameters(self, params):
        """
        Sets the list of search parameters to use.

        :param params: list of AbstractSearchParameter objects
        :type params: list
        """
        array = JavaArray(JavaArray.new_instance("weka.core.setupgenerator.AbstractParameter", len(params)))
        for idx, obj in enumerate(params):
            array[idx] = obj.jobject
        javabridge.call(self.jobject, "setSearchParameters", "([Lweka/core/setupgenerator/AbstractParameter;)V", array.jobject)

    @property
    def best(self):
        """
        Returns the best classifier setup found during the th search.

        :return: the best classifier setup
        :rtype: Classifier
        """
        return Classifier(jobject=javabridge.call(self.jobject, "getBestClassifier", "()Lweka/classifiers/Classifier;"))


class MultipleClassifiersCombiner(Classifier):
    """
    Wrapper class for classifiers that use a multiple base classifiers.
    """

    def __init__(self, classname=None, jobject=None, options=None):
        """
        Initializes the specified classifier using either the classname or the supplied JB_Object.

        :param classname: the classname of the classifier
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: list of commandline options
        :type options: list
        """
        if jobject is None:
            jobject = Classifier.new_instance(classname)
        self.enforce_type(jobject, "weka.classifiers.MultipleClassifiersCombiner")
        super(MultipleClassifiersCombiner, self).__init__(classname=classname, jobject=jobject, options=options)

    @property
    def classifiers(self):
        """
        Returns the list of base classifiers.

        :return: the classifier list
        :rtype: list
        """
        objects = javabridge.get_env().get_object_array_elements(
            javabridge.call(self.jobject, "getClassifiers", "()[Lweka/classifiers/Classifier;"))
        result = []
        for obj in objects:
            result.append(Classifier(jobject=obj))
        return result

    @classifiers.setter
    def classifiers(self, classifiers):
        """
        Sets the base classifiers.

        :param classifiers: the list of base classifiers to use
        :type classifiers: list
        """
        obj = []
        for classifier in classifiers:
            obj.append(classifier.jobject)
        javabridge.call(self.jobject, "setClassifiers", "([Lweka/classifiers/Classifier;)V", obj)

    def clear(self):
        """
        Removes all classifiers.
        """
        self.classifiers = []

    def append(self, classifier):
        """
        Appends the classifier to the current list of classifiers.

        :param classifier: the classifier to add
        :type classifier: Classifier
        """
        l = self.classifiers
        l.append(classifier)
        self.classifiers = l


class Kernel(OptionHandler):
    """
    Wrapper class for kernels.
    """

    def __init__(self, classname=None, jobject=None, options=None):
        """
        Initializes the specified kernel using either the classname or the supplied JB_Object.

        :param classname: the classname of the kernel
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        if jobject is None:
            jobject = Classifier.new_instance(classname)
        self.enforce_type(jobject, "weka.classifiers.functions.supportVector.Kernel")
        super(Kernel, self).__init__(jobject=jobject, options=options)

    def capabilities(self):
        """
        Returns the capabilities of the classifier.

        :return: the capabilities
        :rtype: Capabilities
        """
        return Capabilities(javabridge.call(self.jobject, "getCapabilities", "()Lweka/core/Capabilities;"))

    @property
    def checks_turned_off(self):
        """
        Returns whether checks are turned off.

        :return: True if checks turned off
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getChecksTurnedOff", "()Z")

    @checks_turned_off.setter
    def checks_turned_off(self, off):
        """
        Turns any checks on/off.

        :param off: True to turn off checks
        :type off: bool
        """
        javabridge.call(self.jobject, "setChecksTurnedOff", "(Z)V", off)

    def clean(self):
        """
        Frees the memory used by the kernel.
        """
        javabridge.call(self.jobject, "clean", "()V")

    def build_kernel(self, data):
        """
        Builds the classifier with the data.

        :param data: the data to train the classifier with
        :type data: Instances
        """
        javabridge.call(self.jobject, "buildKernel", "(Lweka/core/Instances;)V", data.jobject)

    def eval(self, id1, id2, inst1):
        """
        Computes the result of the kernel function for two instances. If id1 == -1, eval use inst1 instead of an
        instance in the dataset.

        :param id1: the index of the first instance in the dataset
        :type id1: int
        :param id2: the index of the second instance in the dataset
        :type id2: int
        :param inst1: the instance corresponding to id1 (used if id1 == -1)
        :type inst1: Instance
        """
        jinst1 = None
        if inst1 is not None:
            jinst1 = inst1.jobject
        return javabridge.call(self.jobject, "eval", "(IILweka/core/Instance;)D", id1, id2, jinst1)

    @classmethod
    def make_copy(cls, kernel):
        """
        Creates a copy of the kernel.

        :param kernel: the kernel to copy
        :type kernel: Kernel
        :return: the copy of the kernel
        :rtype: Kernel
        """
        return Kernel(
            jobject=javabridge.static_call(
                "weka/classifiers/functions/supportVector/Kernel", "makeCopy",
                "(Lweka/classifiers/functions/supportVector/Kernel;)Lweka/classifiers/functions/supportVector/Kernel;",
                kernel.jobject))


class KernelClassifier(Classifier):
    """
    Wrapper class for classifiers that have a kernel property, like SMO.
    """

    def __init__(self, classname=None, jobject=None, options=None):
        """
        Initializes the specified classifier using either the classname or the supplied JB_Object.

        :param classname: the classname of the classifier
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: list of commandline options
        :type options: list
        """
        if jobject is None:
            jobject = Classifier.new_instance(classname)
        self.enforce_type(jobject, "weka.classifiers.Classifier")
        if not javabridge.static_call(
                "weka/classifiers/KernelHelper", "hasKernelProperty",
                "(Ljava/lang/Object;)Z",
                jobject):
            raise Exception("Does not handle a kernel: " + classes.get_classname(jobject))
        super(KernelClassifier, self).__init__(classname=classname, jobject=jobject, options=options)

    @property
    def kernel(self):
        """
        Returns the current kernel.

        :return: the kernel or None if none found
        :rtype: Kernel
        """
        result = javabridge.static_call(
            "weka/classifiers/KernelHelper", "getKernel",
            "(Ljava/lang/Object;)Lweka/classifiers/functions/supportVector/Kernel;",
            self.jobject)
        if result is None:
            return None
        else:
            return Kernel(jobject=result)

    @kernel.setter
    def kernel(self, kernel):
        """
        Sets the kernel.

        :param kernel: the kernel to set
        :type kernel: Kernel
        """
        result = javabridge.static_call(
            "weka/classifiers/KernelHelper", "setKernel",
            "(Ljava/lang/Object;Lweka/classifiers/functions/supportVector/Kernel;)Z",
            self.jobject, kernel.jobject)
        if not result:
            raise Exception("Failed to set kernel!")


class Prediction(JavaObject):
    """
    Wrapper class for a prediction.
    """

    def __init__(self, jobject):
        """
        Initializes the wrapper.

        :param jobject: the prediction to wrap
        :type jobject: JB_Object
        """
        self.enforce_type(jobject, "weka.classifiers.evaluation.Prediction")
        super(Prediction, self).__init__(jobject)

    @property
    def actual(self):
        """
        Returns the actual value.

        :return: the actual value (internal representation)
        :rtype: float
        """
        return javabridge.call(self.jobject, "actual", "()D")

    @property
    def predicted(self):
        """
        Returns the predicted value.

        :return: the predicted value (internal representation)
        :rtype: float
        """
        return javabridge.call(self.jobject, "predicted", "()D")

    @property
    def weight(self):
        """
        Returns the weight.

        :return: the weight of the Instance that was used
        :rtype: float
        """
        return javabridge.call(self.jobject, "weight", "()D")


class NominalPrediction(Prediction):
    """
    Wrapper class for a nominal prediction.
    """

    def __init__(self, jobject):
        """
        Initializes the wrapper.

        :param jobject: the prediction to wrap
        :type jobject: JB_Object
        """
        self.enforce_type(jobject, "weka.classifiers.evaluation.NominalPrediction")
        super(NominalPrediction, self).__init__(jobject)

    @property
    def distribution(self):
        """
        Returns the class distribution.

        :return: the class distribution list
        :rtype: ndarray
        """
        return javabridge.get_env().get_double_array_elements(javabridge.call(self.jobject, "distribution", "()[D"))

    @property
    def margin(self):
        """
        Returns the margin.

        :return: the margin
        :rtype: float
        """
        return javabridge.call(self.jobject, "margin", "()D")


class NumericPrediction(Prediction):
    """
    Wrapper class for a numeric prediction.
    """

    def __init__(self, jobject):
        """
        Initializes the wrapper.

        :param jobject: the prediction to wrap
        :type jobject: JB_Object
        """
        self.enforce_type(jobject, "weka.classifiers.evaluation.NumericPrediction")
        super(NumericPrediction, self).__init__(jobject)

    @property
    def error(self):
        """
        Returns the error.

        :return: the error
        :rtype: float
        """
        return javabridge.call(self.jobject, "error", "()D")

    @property
    def prediction_intervals(self):
        """
        Returns the prediction intervals.

        :return: the intervals
        :rtype: ndarray
        """
        return typeconv.jdouble_matrix_to_ndarray(javabridge.call(self.jobject, "predictionIntervals", "()[[D"))


class CostMatrix(JavaObject):
    """
    Class for storing and manipulating a misclassification cost matrix. The element at position i,j in the matrix
    is the penalty for classifying an instance of class j as class i. Cost values can be fixed or computed on a
    per-instance basis (cost sensitive evaluation only) from the value of an attribute or an expression involving
    attribute(s).
    """

    def __init__(self, matrx=None, num_classes=None):
        """
        Initializes the matrix object.

        :param matrx: the matrix to copy
        :type matrx: CostMatrix or ndarray or JB_Object
        :param num_classes: the number of classes
        :type num_classes: int
        """
        if matrx is not None:
            if isinstance(matrx, CostMatrix):
                jobject = javabridge.make_instance(
                    "weka/classifiers/CostMatrix", "(Lweka/classifiers/CostMatrix;)V", matrx.jobject)
                super(CostMatrix, self).__init__(jobject)
            elif isinstance(matrx, ndarray):
                shp = matrx.shape
                if len(shp) != 2:
                    raise Exception("Numpy array must be a 2-dimensional array!")
                rows, cols = shp
                if rows == cols:
                    cmatrix = CostMatrix(num_classes=rows)
                    for r in range(rows):
                        for c in range(cols):
                            cmatrix.set_element(r, c, matrx[r][c])
                    super(CostMatrix, self).__init__(cmatrix.jobject)
                else:
                    raise Exception("Numpy array must be a square matrix!")
            elif isinstance(matrx, javabridge.JB_Object):
                super(CostMatrix, self).__init__(matrx)
            else:
                raise Exception(
                    "Matrix must be either a CostMatrix or a 2-dimensional numpy array: " + str(type(matrx)))
        elif not num_classes is None:
            jobject = javabridge.make_instance(
                "weka/classifiers/CostMatrix", "(I)V", num_classes)
            super(CostMatrix, self).__init__(jobject)
        else:
            raise Exception("Either matrix or number of classes must be provided!")

    def apply_cost_matrix(self, data, rnd):
        """
        Applies the cost matrix to the data.

        :param data: the data to apply to
        :type data: Instances
        :param rnd: the random number generator
        :type rnd: Random
        """
        return Instances(
            javabridge.call(
                self.jobject, "applyCostMatrix", "(Lweka/core/Instances;Ljava/util/Random;)Lweka/core/Instances;",
                data.jobject, rnd.jobject))

    def expected_costs(self, class_probs, inst=None):
        """
        Calculates the expected misclassification cost for each possible class value, given class probability
        estimates.

        :param class_probs: the class probabilities
        :type class_probs: ndarray
        :return: the calculated costs
        :rtype: ndarray
        """
        if inst is None:
            costs = javabridge.call(
                self.jobject, "expectedCosts", "([D)[D", javabridge.get_env().make_double_array(class_probs))
            return javabridge.get_env().get_double_array_elements(costs)
        else:
            costs = javabridge.call(
                self.jobject, "expectedCosts", "([DLweka/core/Instance;)[D",
                javabridge.get_env().make_double_array(class_probs), inst.jobject)
            return javabridge.get_env().get_double_array_elements(costs)

    def get_cell(self, row, col):
        """
        Returns the JB_Object at the specified location.

        :param row: the 0-based index of the row
        :type row: int
        :param col: the 0-based index of the column
        :type col: int
        :return: the object in that cell
        :rtype: JB_Object
        """
        return javabridge.call(
            self.jobject, "getCell", "(II)Ljava/lang/Object;", row, col)

    def set_cell(self, row, col, obj):
        """
        Sets the JB_Object at the specified location. Automatically unwraps JavaObject.

        :param row: the 0-based index of the row
        :type row: int
        :param col: the 0-based index of the column
        :type col: int
        :param obj: the object for that cell
        :type obj: object
        """
        if isinstance(obj, JavaObject):
            obj = obj.jobject
        javabridge.call(
            self.jobject, "setCell", "(IILjava/lang/Object;)V", row, col, obj)

    def get_element(self, row, col, inst=None):
        """
        Returns the value at the specified location.

        :param row: the 0-based index of the row
        :type row: int
        :param col: the 0-based index of the column
        :type col: int
        :param inst: the Instace
        :type inst: Instance
        :return: the value in that cell
        :rtype: float
        """
        if inst is None:
            return javabridge.call(
                self.jobject, "getElement", "(II)D", row, col)
        else:
            return javabridge.call(
                self.jobject, "getElement", "(IILweka/core/Instance;)D", row, col, inst.jobject)

    def set_element(self, row, col, value):
        """
        Sets the float value at the specified location.

        :param row: the 0-based index of the row
        :type row: int
        :param col: the 0-based index of the column
        :type col: int
        :param value: the float value for that cell
        :type value: float
        """
        javabridge.call(
            self.jobject, "setElement", "(IID)V", row, col, value)

    def get_max_cost(self, class_value, inst=None):
        """
        Gets the maximum cost for a particular class value.

        :param class_value: the class value to get the maximum cost for
        :type class_value: int
        :param inst: the Instance
        :type inst: Instance
        :return: the cost
        :rtype: float
        """
        if inst is None:
            return javabridge.call(
                self.jobject, "getMaxCost", "(I)D", class_value)
        else:
            return javabridge.call(
                self.jobject, "getElement", "(ILweka/core/Instance;)D", class_value, inst.jobject)

    def initialize(self):
        """
        Initializes the matrix.
        """
        javabridge.call(self.jobject, "initialize", "()V")

    def normalize(self):
        """
        Normalizes the matrix.
        """
        javabridge.call(self.jobject, "normalize", "()V")

    @property
    def num_columns(self):
        """
        Returns the number of columns.

        :return: the number of columns
        :rtype: int
        """
        return javabridge.call(self.jobject, "numColumns", "()I")

    @property
    def num_rows(self):
        """
        Returns the number of rows.

        :return: the number of rows
        :rtype: int
        """
        return javabridge.call(self.jobject, "numRows", "()I")

    @property
    def size(self):
        """
        Returns the number of rows/columns.

        :return: the number of rows/columns
        :rtype: int
        """
        return javabridge.call(self.jobject, "size", "()I")

    def to_matlab(self):
        """
        Returns the matrix in Matlab format.

        :return: the matrix as Matlab formatted string
        :rtype: str
        """
        return javabridge.call(self.jobject, "toMatlab", "()Ljava/lang/String;")

    @classmethod
    def parse_matlab(cls, matlab):
        """
        Parses the costmatrix definition in matlab format and returns a matrix.

        :param matlab: the matlab matrix string, eg [1 2; 3 4].
        :type matlab: str
        :return: the generated matrix
        :rtype: CostMatrix
        """
        return CostMatrix(
            matrx=javabridge.static_call(
                "weka/classifiers/CostMatrix", "parseMatlab",
                "(Ljava/lang/String;)Lweka/classifiers/CostMatrix;", matlab))


class Evaluation(JavaObject):
    """
    Evaluation class for classifiers.
    """

    def __init__(self, data, cost_matrix=None):
        """
        Initializes an Evaluation object.

        :param data: the data to use to initialize the priors with
        :type data: Instances
        :param cost_matrix: the cost matrix to use for initializing
        :type cost_matrix: CostMatrix
        """
        if cost_matrix is None:
            jobject = javabridge.make_instance(
                "weka/classifiers/EvaluationWrapper", "(Lweka/core/Instances;)V",
                data.jobject)
        else:
            jobject = javabridge.make_instance(
                "weka/classifiers/EvaluationWrapper", "(Lweka/core/Instances;Lweka/classifiers/CostMatrix;)V",
                data.jobject, cost_matrix.jobject)
        self.wrapper = jobject
        jobject = javabridge.call(jobject, "getEvaluation", "()Lweka/classifiers/Evaluation;")
        super(Evaluation, self).__init__(jobject)

    def crossvalidate_model(self, classifier, data, num_folds, rnd, output=None):
        """
        Crossvalidates the model using the specified data, number of folds and random number generator wrapper.

        :param classifier: the classifier to cross-validate
        :type classifier: Classifier
        :param data: the data to evaluate on
        :type data: Instances
        :param num_folds: the number of folds
        :type num_folds: int
        :param rnd: the random number generator to use
        :type rnd: Random
        :param output: the output generator to use
        :type output: PredictionOutput
        """
        if output is None:
            generator = []
        else:
            generator = [output.jobject]
        javabridge.call(
            self.jobject, "crossValidateModel",
            "(Lweka/classifiers/Classifier;Lweka/core/Instances;ILjava/util/Random;[Ljava/lang/Object;)V",
            classifier.jobject, data.jobject, num_folds, rnd.jobject, generator)

    def evaluate_train_test_split(self, classifier, data, percentage, rnd=None, output=None):
        """
        Splits the data into train and test, builds the classifier with the training data and
        evaluates it against the test set.

        :param classifier: the classifier to cross-validate
        :type classifier: Classifier
        :param data: the data to evaluate on
        :type data: Instances
        :param percentage: the percentage split to use (amount to use for training)
        :type percentage: double
        :param rnd: the random number generator to use, if None the order gets preserved
        :type rnd: Random
        :param output: the output generator to use
        :type output: PredictionOutput
        """
        train_inst, test_inst = data.train_test_split(percentage, rnd=rnd)
        cls = Classifier.make_copy(classifier)
        cls.build_classifier(train_inst)
        self.test_model(cls, test_inst, output=output)

    def test_model(self, classifier, data, output=None):
        """
        Evaluates the built model using the specified test data and returns the classifications.

        :param classifier: the trained classifier to evaluate
        :type classifier: Classifier
        :param data: the data to evaluate on
        :type data: Instances
        :param output: the output generator to use
        :type output: PredictionOutput
        :return: the classifications
        :rtype: ndarray
        """
        if output is None:
            generator = []
        else:
            output.header = data
            generator = [output.jobject]
        cls = javabridge.call(
            self.jobject, "evaluateModel",
            "(Lweka/classifiers/Classifier;Lweka/core/Instances;[Ljava/lang/Object;)[D",
            classifier.jobject, data.jobject, generator)
        if cls is None:
            return None
        else:
            return javabridge.get_env().get_double_array_elements(cls)

    def test_model_once(self, classifier, inst):
        """
        Evaluates the built model using the specified test instance and returns the classification.

        :param classifier: the classifier to cross-validate
        :type classifier: Classifier
        :param inst: the Instance to evaluate on
        :type inst: Instances
        :return: the classification
        :rtype: float
        """
        return javabridge.call(
            self.jobject, "evaluateModelOnce",
            "(Lweka/classifiers/Classifier;Lweka/core/Instance;)D",
            classifier.jobject, inst.jobject)

    def summary(self, title=None, complexity=False):
        """
        Generates a summary.

        :param title: optional title
        :type title: str
        :param complexity: whether to print the complexity information as well
        :type complexity: bool
        :return: the summary
        :rtype: str
        """
        if title is None:
            return javabridge.call(
                self.jobject, "toSummaryString", "()Ljava/lang/String;")
        else:
            return javabridge.call(
                self.jobject, "toSummaryString", "(Ljava/lang/String;Z)Ljava/lang/String;", title, complexity)

    def class_details(self, title=None):
        """
        Generates the class details.

        :param title: optional title
        :type title: str
        :return: the details
        :rtype: str
        """
        if title is None:
            return javabridge.call(
                self.jobject, "toClassDetailsString", "()Ljava/lang/String;")
        else:
            return javabridge.call(
                self.jobject, "toClassDetailsString", "(Ljava/lang/String;)Ljava/lang/String;", title)

    def matrix(self, title=None):
        """
        Generates the confusion matrix.

        :param title: optional title
        :type title: str
        :return: the matrix
        :rtype: str
        """
        if title is None:
            return javabridge.call(self.jobject, "toMatrixString", "()Ljava/lang/String;")
        else:
            return javabridge.call(self.jobject, "toMatrixString", "(Ljava/lang/String;)Ljava/lang/String;", title)

    def cumulative_margin_distribution(self):
        """
        Output the cumulative margin distribution as a string suitable for input for gnuplot or similar package.

        :return: the cumulative margin distribution
        :rtype: str
        """
        return javabridge.call(self.jobject, "toCumulativeMarginDistributionString", "()Ljava/lang/String;")

    def area_under_prc(self, class_index):
        """
        Returns the area under precision recall curve.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the area
        :rtype: float
        """
        return javabridge.call(self.jobject, "areaUnderPRC", "(I)D", class_index)

    @property
    def weighted_area_under_prc(self):
        """
        Returns the weighted area under precision recall curve.

        :return: the weighted area
        :rtype: float
        """
        return javabridge.call(self.jobject, "weightedAreaUnderPRC", "()D")

    def area_under_roc(self, class_index):
        """
        Returns the area under receiver operators characteristics curve.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the area
        :rtype: float
        """
        return javabridge.call(self.jobject, "areaUnderROC", "(I)D", class_index)

    @property
    def weighted_area_under_roc(self):
        """
        Returns the weighted area under receiver operator characteristic curve.

        :return: the weighted area
        :rtype: float
        """
        return javabridge.call(self.jobject, "weightedAreaUnderROC", "()D")

    @property
    def avg_cost(self):
        """
        Returns the average cost.

        :return: the cost
        :rtype: float
        """
        return javabridge.call(self.jobject, "avgCost", "()D")

    @property
    def total_cost(self):
        """
        Returns the total cost.

        :return: the cost
        :rtype: float
        """
        return javabridge.call(self.jobject, "totalCost", "()D")

    @property
    def confusion_matrix(self):
        """
        Returns the confusion matrix.

        :return: the matrix
        :rtype: ndarray
        """
        return typeconv.jdouble_matrix_to_ndarray(javabridge.call(self.jobject, "confusionMatrix", "()[[D"))

    @property
    def correct(self):
        """
        Returns the correct count (nominal classes).

        :return: the count
        :rtype: float
        """
        return javabridge.call(self.jobject, "correct", "()D")

    @property
    def incorrect(self):
        """
        Returns the incorrect count (nominal classes).

        :return: the count
        :rtype: float
        """
        return javabridge.call(self.jobject, "incorrect", "()D")

    @property
    def unclassified(self):
        """
        Returns the unclassified count.

        :return: the count
        :rtype: float
        """
        return javabridge.call(self.jobject, "unclassified", "()D")

    @property
    def num_instances(self):
        """
        Returns the number of instances that had a known class value.

        :return: the number of instances
        :rtype: float
        """
        return javabridge.call(self.jobject, "numInstances", "()D")

    @property
    def percent_correct(self):
        """
        Returns the percent correct (nominal classes).

        :return: the percentage
        :rtype: float
        """
        return javabridge.call(self.jobject, "pctCorrect", "()D")

    @property
    def percent_incorrect(self):
        """
        Returns the percent incorrect (nominal classes).

        :return: the percentage
        :rtype: float
        """
        return javabridge.call(self.jobject, "pctIncorrect", "()D")

    @property
    def percent_unclassified(self):
        """
        Returns the percent unclassified.

        :return: the percentage
        :rtype: float
        """
        return javabridge.call(self.jobject, "pctUnclassified", "()D")

    @property
    def correlation_coefficient(self):
        """
        Returns the correlation coefficient (numeric classes).

        :return: the coefficient
        :rtype: float
        """
        return javabridge.call(self.jobject, "correlationCoefficient", "()D")

    def matthews_correlation_coefficient(self, class_index):
        """
        Returns the Matthews correlation coefficient (nominal classes).

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the coefficient
        :rtype: float
        """
        return javabridge.call(self.jobject, "matthewsCorrelationCoefficient", "(I)D", class_index)

    @property
    def weighted_matthews_correlation(self):
        """
        Returns the weighted Matthews correlation (nominal classes).

        :return: the correlation
        :rtype: float
        """
        return javabridge.call(self.jobject, "weightedMatthewsCorrelation", "()D")

    @property
    def coverage_of_test_cases_by_predicted_regions(self):
        """
        Returns the coverage of the test cases by the predicted regions at the confidence level
        specified when evaluation was performed.

        :return: the coverage
        :rtype: float
        """
        return javabridge.call(self.jobject, "coverageOfTestCasesByPredictedRegions", "()D")

    @property
    def size_of_predicted_regions(self):
        """
        Returns  the average size of the predicted regions, relative to the range of the target in the
        training data, at the confidence level specified when evaluation was performed.

        :return:the size of the regions
        :rtype: float
        """
        return javabridge.call(self.jobject, "sizeOfPredictedRegions", "()D")

    @property
    def error_rate(self):
        """
        Returns the error rate (numeric classes).

        :return: the rate
        :rtype: float
        """
        return javabridge.call(self.jobject, "errorRate", "()D")

    @property
    def mean_absolute_error(self):
        """
        Returns the mean absolute error.

        :return: the error
        :rtype: float
        """
        return javabridge.call(self.jobject, "meanAbsoluteError", "()D")

    @property
    def relative_absolute_error(self):
        """
        Returns the relative absolute error.

        :return: the error
        :rtype: float
        """
        return javabridge.call(self.jobject, "relativeAbsoluteError", "()D")

    @property
    def root_mean_squared_error(self):
        """
        Returns the root mean squared error.

        :return: the error
        :rtype: float
        """
        return javabridge.call(self.jobject, "rootMeanSquaredError", "()D")

    @property
    def root_relative_squared_error(self):
        """
        Returns the root relative squared error.

        :return: the error
        :rtype: float
        """
        return javabridge.call(self.jobject, "rootRelativeSquaredError", "()D")

    @property
    def root_mean_prior_squared_error(self):
        """
        Returns the root mean prior squared error.

        :return: the error
        :rtype: float
        """
        return javabridge.call(self.jobject, "rootMeanPriorSquaredError", "()D")

    @property
    def mean_prior_absolute_error(self):
        """
        Returns the mean prior absolute error.

        :return: the error
        :rtype: float
        """
        return javabridge.call(self.jobject, "meanPriorAbsoluteError", "()D")

    def false_negative_rate(self, class_index):
        """
        Returns the false negative rate.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the rate
        :rtype: float
        """
        return javabridge.call(self.jobject, "falseNegativeRate", "(I)D", class_index)

    @property
    def weighted_false_negative_rate(self):
        """
        Returns the weighted false negative rate.

        :return: the rate
        :rtype: float
        """
        return javabridge.call(self.jobject, "weightedFalseNegativeRate", "()D")

    def false_positive_rate(self, class_index):
        """
        Returns the false positive rate.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the rate
        :rtype: float
        """
        return javabridge.call(self.jobject, "falsePositiveRate", "(I)D", class_index)

    @property
    def weighted_false_positive_rate(self):
        """
        Returns the weighted false positive rate.

        :return: the rate
        :rtype: float
        """
        return javabridge.call(self.jobject, "weightedFalsePositiveRate", "()D")

    def num_false_negatives(self, class_index):
        """
        Returns the number of false negatives.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the count
        :rtype: float
        """
        return javabridge.call(self.jobject, "numFalseNegatives", "(I)D", class_index)

    def true_negative_rate(self, class_index):
        """
        Returns the true negative rate.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the rate
        :rtype: float
        """
        return javabridge.call(self.jobject, "trueNegativeRate", "(I)D", class_index)

    @property
    def weighted_true_negative_rate(self):
        """
        Returns the weighted true negative rate.

        :return: the rate
        :rtype: float
        """
        return javabridge.call(self.jobject, "weightedTrueNegativeRate", "()D")

    def num_true_negatives(self, class_index):
        """
        Returns the number of true negatives.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the count
        :rtype: float
        """
        return javabridge.call(self.jobject, "numTrueNegatives", "(I)D", class_index)

    def num_false_positives(self, class_index):
        """
        Returns the number of false positives.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the count
        :rtype: float
        """
        return javabridge.call(self.jobject, "numFalsePositives", "(I)D", class_index)

    def true_positive_rate(self, class_index):
        """
        Returns the true positive rate.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the rate
        :rtype: float
        """
        return javabridge.call(self.jobject, "truePositiveRate", "(I)D", class_index)

    @property
    def weighted_true_positive_rate(self):
        """
        Returns the weighted true positive rate.

        :return: the rate
        :rtype: float
        """
        return javabridge.call(self.jobject, "weightedTruePositiveRate", "()D")

    def num_true_positives(self, class_index):
        """
        Returns the number of true positives.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the count
        :rtype: float
        """
        return javabridge.call(self.jobject, "numTruePositives", "(I)D", class_index)

    def f_measure(self, class_index):
        """
        Returns the f measure.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the measure
        :rtype: float
        """
        return javabridge.call(self.jobject, "fMeasure", "(I)D", class_index)

    @property
    def weighted_f_measure(self):
        """
        Returns the weighted f measure.

        :return: the measure
        :rtype: float
        """
        return javabridge.call(self.jobject, "weightedFMeasure", "()D")

    @property
    def unweighted_macro_f_measure(self):
        """
        Returns the unweighted macro-averaged F-measure.

        :return: the measure
        :rtype: float
        """
        return javabridge.call(self.jobject, "unweightedMacroFmeasure", "()D")

    @property
    def unweighted_micro_f_measure(self):
        """
        Returns the unweighted micro-averaged F-measure.

        :return: the measure
        :rtype: float
        """
        return javabridge.call(self.jobject, "unweightedMicroFmeasure", "()D")

    def precision(self, class_index):
        """
        Returns the precision.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the precision
        :rtype: float
        """
        return javabridge.call(self.jobject, "precision", "(I)D", class_index)

    @property
    def weighted_precision(self):
        """
        Returns the weighted precision.

        :return: the precision
        :rtype: float
        """
        return javabridge.call(self.jobject, "weightedPrecision", "()D")

    def recall(self, class_index):
        """
        Returns the recall.

        :param class_index: the 0-based index of the class label
        :type class_index: int
        :return: the recall
        :rtype: float
        """
        return javabridge.call(self.jobject, "recall", "(I)D", class_index)

    @property
    def weighted_recall(self):
        """
        Returns the weighted recall.

        :return: the recall
        :rtype: float
        """
        return javabridge.call(self.jobject, "weightedRecall", "()D")

    @property
    def kappa(self):
        """
        Returns kappa.

        :return: kappa
        :rtype: float
        """
        return javabridge.call(self.jobject, "kappa", "()D")

    @property
    def kb_information(self):
        """
        Returns KB information.

        :return: the information
        :rtype: float
        """
        return javabridge.call(self.jobject, "KBInformation", "()D")

    @property
    def kb_mean_information(self):
        """
        Returns KB mean information.

        :return: the information
        :rtype: float
        """
        return javabridge.call(self.jobject, "KBMeanInformation", "()D")

    @property
    def kb_relative_information(self):
        """
        Returns KB relative information.

        :return: the information
        :rtype: float
        """
        return javabridge.call(self.jobject, "KBRelativeInformation", "()D")

    @property
    def sf_entropy_gain(self):
        """
        Returns the total SF, which is the null model entropy minus the scheme entropy.

        :return: the gain
        :rtype: float
        """
        return javabridge.call(self.jobject, "SFEntropyGain", "()D")

    @property
    def sf_mean_entropy_gain(self):
        """
        Returns the SF per instance, which is the null model entropy minus the scheme entropy, per instance.

        :return: the gain
        :rtype: float
        """
        return javabridge.call(self.jobject, "SFMeanEntropyGain", "()D")

    @property
    def sf_mean_prior_entropy(self):
        """
        Returns the entropy per instance for the null model.

        :return: the entropy
        :rtype: float
        """
        return javabridge.call(self.jobject, "SFMeanPriorEntropy", "()D")

    @property
    def sf_mean_scheme_entropy(self):
        """
        Returns the entropy per instance for the scheme.

        :return: the entropy
        :rtype: float
        """
        return javabridge.call(self.jobject, "SFMeanSchemeEntropy", "()D")

    @property
    def sf_prior_entropy(self):
        """
        Returns the total entropy for the null model.

        :return: the entropy
        :rtype: float
        """
        return javabridge.call(self.jobject, "SFPriorEntropy", "()D")

    @property
    def sf_scheme_entropy(self):
        """
        Returns the total entropy for the scheme.

        :return: the entropy
        :rtype: float
        """
        return javabridge.call(self.jobject, "SFchemeEntropy", "()D")

    @property
    def class_priors(self):
        """
        Returns the class priors.

        :return: the priors
        :rtype: ndarray
        """
        return javabridge.get_env().get_double_array_elements(javabridge.call(self.jobject, "getClassPriors", "()[D"))

    @class_priors.setter
    def class_priors(self, data):
        """
        Sets the class priors derived from the dataset.

        :param data: the dataset to derive the priors from
        :type data: Instances
        """
        javabridge.call(self.jobject, "setClassPriors", "(Lweka/core/Instances;)V", data)

    @property
    def header(self):
        """
        Returns the header format.

        :return: the header format
        :rtype: Instances
        """
        return Instances(javabridge.call(self.jobject, "getHeader", "()Lweka/core/Instances;"))

    @property
    def discard_predictions(self):
        """
        Returns whether to discard predictions (saves memory).

        :return: True if to discard
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getDiscardPredictions", "()Z")

    @discard_predictions.setter
    def discard_predictions(self, discard):
        """
        Sets whether to discard predictions (saves memory).

        :param discard: True if to discard predictions
        :type discard: bool
        """
        javabridge.call(self.jobject, "setDiscardPredictions", "(Z)V", discard)

    @property
    def predictions(self):
        """
        Returns the predictions.

        :return: the predictions. None if not available
        :rtype: list
        """
        preds = javabridge.get_collection_wrapper(
            javabridge.call(self.jobject, "predictions", "()Ljava/util/ArrayList;"))
        if self.discard_predictions:
            result = None
        else:
            result = []
            for pred in preds:
                if is_instance_of(pred, "weka.classifiers.evaluation.NominalPrediction"):
                    result.append(NominalPrediction(pred))
                elif is_instance_of(pred, "weka.classifiers.evaluation.NumericPrediction"):
                    result.append(NumericPrediction(pred))
                else:
                    result.append(Prediction(pred))
        return result

    @classmethod
    def evaluate_model(cls, classifier, args):
        """
        Evaluates the classifier with the given options.

        :param classifier: the classifier instance to use
        :type classifier: Classifier
        :param args: the command-line arguments to use
        :type args: list
        :return: the evaluation string
        :rtype: str
        """
        return javabridge.static_call(
            "Lweka/classifiers/Evaluation;", "evaluateModel",
            "(Lweka/classifiers/Classifier;[Ljava/lang/String;)Ljava/lang/String;",
            classifier.jobject, args)


class PredictionOutput(OptionHandler):
    """
    For collecting predictions and generating output from.
    Must be derived from weka.classifiers.evaluation.output.prediction.AbstractOutput
    """

    def __init__(self, classname="weka.classifiers.evaluation.output.prediction.PlainText", jobject=None, options=None):
        """
        Initializes the specified output generator using either the classname or the supplied JB_Object.

        :param classname: the classname of the generator
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        if jobject is None:
            jobject = PredictionOutput.new_instance(classname)
        self.enforce_type(jobject, "weka.classifiers.evaluation.output.prediction.AbstractOutput")
        super(PredictionOutput, self).__init__(jobject=jobject, options=options)
        buf = javabridge.make_instance("java/lang/StringBuffer", "()V")
        javabridge.call(self.jobject, "setBuffer", "(Ljava/lang/StringBuffer;)V", buf)

    @property
    def header(self):
        """
        Returns the header format.

        :return: The dataset format
        :rtype: Instances
        """
        return Instances(javabridge.call(self.jobject, "getHeader", "()Lweka/core/Instances;"))

    @header.setter
    def header(self, data):
        """
        Sets the header format.

        :param data: The dataset format
        :type data: Instances
        """
        javabridge.call(self.jobject, "setHeader", "(Lweka/core/Instances;)V", data.jobject)

    def print_header(self):
        """
        Prints the header to the buffer.
        """
        javabridge.call(self.jobject, "printHeader", "()V")

    def print_footer(self):
        """
        Prints the footer to the buffer.
        """
        javabridge.call(self.jobject, "printFooter", "()V")

    def print_all(self, cls, data):
        """
        Prints the header, classifications and footer to the buffer.

        :param cls: the classifier
        :type cls: Classifier
        :param data: the test data
        :type data: Instances
        """
        javabridge.call(
            self.jobject, "print", "(Lweka/classifiers/Classifier;Lweka/core/Instances;)V",
            cls.jobject, data.jobject)

    def print_classifications(self, cls, data):
        """
        Prints the classifications to the buffer.

        :param cls: the classifier
        :type cls: Classifier
        :param data: the test data
        :type data: Instances
        """
        javabridge.call(
            self.jobject, "printClassifications", "(Lweka/classifiers/Classifier;Lweka/core/Instances;)V",
            cls.jobject, data.jobject)

    def print_classification(self, cls, inst, index):
        """
        Prints the classification to the buffer.

        :param cls: the classifier
        :type cls: Classifier
        :param inst: the test instance
        :type inst: Instance
        :param index: the 0-based index of the test instance
        :type index: int
        """
        javabridge.call(
            self.jobject, "printClassification", "(Lweka/classifiers/Classifier;Lweka/core/Instance;I)V",
            cls.jobject, inst.jobject, index)

    def buffer_content(self):
        """
        Returns the content of the buffer as string.

        :return: The buffer content
        :rtype: str
        """
        return javabridge.to_string(javabridge.call(self.jobject, "getBuffer", "()Ljava/lang/StringBuffer;"))

    def __str__(self):
        """
        Returns the content of the buffer.

        :return: the current buffer content
        :rtype: str
        """
        return self.buffer_content()


def predictions_to_instances(data, preds):
    """
    Turns the predictions turned into an Instances object.

    :param data: the original dataset format
    :type data: Instances
    :param preds: the predictions to convert
    :type preds: list
    :return: the predictions, None if no predictions present
    :rtype: Instances
    """
    if len(preds) == 0:
        return None

    is_numeric = isinstance(preds[0], NumericPrediction)

    # create header
    atts = []
    if is_numeric:
        atts.append(Attribute.create_numeric("index"))
        atts.append(Attribute.create_numeric("weight"))
        atts.append(Attribute.create_numeric("actual"))
        atts.append(Attribute.create_numeric("predicted"))
        atts.append(Attribute.create_numeric("error"))
    else:
        atts.append(Attribute.create_numeric("index"))
        atts.append(Attribute.create_numeric("weight"))
        atts.append(data.class_attribute.copy(name="actual"))
        atts.append(data.class_attribute.copy(name="predicted"))
        atts.append(Attribute.create_nominal("error", ["no", "yes"]))
        atts.append(Attribute.create_numeric("classification"))
        for i in range(data.class_attribute.num_values):
            atts.append(Attribute.create_numeric("distribution-" + data.class_attribute.value(i)))

    result = Instances.create_instances("Predictions", atts, len(preds))

    count = 0
    for pred in preds:
        count += 1
        if is_numeric:
            values = array([count, pred.weight, pred.actual, pred.predicted, pred.error])
        else:
            if pred.actual == pred.predicted:
                error = 0.0
            else:
                error = 1.0
            l = [count, pred.weight, pred.actual, pred.predicted, error, max(pred.distribution)]
            for i in range(data.class_attribute.num_values):
                l.append(pred.distribution[i])
            values = array(l)
        inst = Instance.create_instance(values)
        result.add_instance(inst)

    return result


def main(args=None):
    """
    Runs a classifier from the command-line. Calls JVM start/stop automatically.
    Use -h to see all options.

    :param args: the command-line arguments to use, uses sys.argv if None
    :type args: list
    """

    parser = argparse.ArgumentParser(
        description='Performs classification/regression from the command-line. Calls JVM start/stop automatically.')
    parser.add_argument("-j", metavar="classpath", dest="classpath", help="additional classpath, jars/directories")
    parser.add_argument("-X", metavar="heap", dest="heap", help="max heap size for jvm, e.g., 512m")
    parser.add_argument("-t", metavar="train", dest="train", required=True, help="Training set file")
    parser.add_argument("-T", metavar="test", dest="test", help="Test set file")
    parser.add_argument("-c", metavar="class index", dest="classindex", help="1-based class attribute index")
    parser.add_argument("-d", metavar="outmodel", dest="outmodel", help="model output file name")
    parser.add_argument("-l", metavar="inmodel", dest="inmodel", help="model input file name")
    parser.add_argument("-x", metavar="num folds", dest="numfolds", help="number of folds for cross-validation")
    parser.add_argument("-s", metavar="seed", dest="seed", help="seed value for randomization")
    parser.add_argument("-v", action="store_true", dest="notrainstats", help="no statistics for training")
    parser.add_argument("-o", action="store_true", dest="onlystats", help="only statistics, don't output model")
    parser.add_argument("-i", action="store_true", dest="irstats", help="output information retrieval statistics")
    parser.add_argument("-k", action="store_true", dest="itstats", help="output information theoretic statistics")
    parser.add_argument("-m", metavar="costmatrix", dest="costmatrix", help="cost matrix file")
    parser.add_argument("-g", metavar="graph", dest="graph", help="output file for graph (if supported)")
    parser.add_argument("classifier", help="classifier classname, e.g., weka.classifiers.trees.J48")
    parser.add_argument("option", nargs=argparse.REMAINDER, help="additional classifier options")
    parsed = parser.parse_args(args=args)
    jars = []
    if parsed.classpath is not None:
        jars = parsed.classpath.split(os.pathsep)
    params = []
    if parsed.train is not None:
        params.extend(["-t", parsed.train])
    if parsed.test is not None:
        params.extend(["-T", parsed.test])
    if parsed.classindex is not None:
        params.extend(["-c", parsed.classindex])
    if parsed.outmodel is not None:
        params.extend(["-d", parsed.outmodel])
    if parsed.inmodel is not None:
        params.extend(["-l", parsed.inmodel])
    if parsed.numfolds is not None:
        params.extend(["-x", parsed.numfolds])
    if parsed.seed is not None:
        params.extend(["-s", parsed.seed])
    if parsed.notrainstats:
        params.append("-v")
    if parsed.onlystats:
        params.append("-o")
    if parsed.irstats:
        params.append("-i")
    if parsed.itstats:
        params.append("-k")
    if parsed.costmatrix is not None:
        params.extend(["-m", parsed.costmatrix])
    if parsed.graph is not None:
        params.extend(["-g", parsed.graph])

    jvm.start(jars, max_heap_size=parsed.heap, packages=True)

    logger.debug("Commandline: " + join_options(sys.argv[1:]))

    try:
        classifier = Classifier(classname=parsed.classifier)
        if len(parsed.option) > 0:
            classifier.options = parsed.option
        print(Evaluation.evaluate_model(classifier, params))
    except Exception:
        print(traceback.format_exc())
    finally:
        jvm.stop()


def sys_main():
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    :rtype: int
    """

    try:
        main()
        return 0
    except Exception:
        print(traceback.format_exc())
        return 1


if __name__ == "__main__":
    try:
        main()
    except Exception:
        print(traceback.format_exc())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# control.py
# Copyright (C) 2015-2016 Fracpete (pythonwekawrapper at gmail dot com)


import weka.flow.base as base
from weka.flow.base import Actor, InputConsumer, OutputProducer, Stoppable, StorageHandler, Token
from weka.flow.transformer import Transformer
import weka.core.classes as classes


class ActorHandler(Actor):
    """
    The ancestor for all actors that handle other actors.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the actor handler.

        :param name: the name of the actor handler
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(ActorHandler, self).__init__(name=name, config=config)
        self._director = self.new_director()
        if not classes.has_dict_handler("ActorHandler"):
            classes.register_dict_handler("ActorHandler", ActorHandler.from_dict)

    def new_director(self):
        """
        Creates the director to use for handling the sub-actors.

        :return: the director instance
        :rtype: Director
        """
        raise Exception("Not implemented!")

    def default_actors(self):
        """
        Returns the default actors to use.

        :return: the default actors, if any
        :rtype: list
        """
        return []

    def check_actors(self, actors):
        """
        Performs checks on the actors that are to be used. Raises an exception if invalid setup.

        :param actors: the actors to check
        :type actors: list
        """
        pass

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(ActorHandler, self).fix_config(options)

        opt = "actors"
        if opt not in options:
            options[opt] = self.default_actors()
        if opt not in self.help:
            self.help[opt] = "The list of sub-actors that this actor manages."

        return options

    def to_dict(self):
        """
        Returns a dictionary that represents this object, to be used for JSONification.

        :return: the object dictionary
        :rtype: dict
        """
        result = super(ActorHandler, self).to_dict()
        result["type"] = "ActorHandler"
        del result["config"]["actors"]
        result["actors"] = []
        for actor in self.actors:
            result["actors"].append(actor.to_dict())
        return result

    @classmethod
    def from_dict(cls, d):
        """
        Restores an object state from a dictionary, used in de-JSONification.

        :param d: the object dictionary
        :type d: dict
        :return: the object
        :rtype: object
        """
        result = super(ActorHandler, cls).from_dict(d)
        if "actors" in d:
            l = d["actors"]
            for e in l:
                if u"type" in e:
                    typestr = e[u"type"]
                else:
                    typestr = e["type"]
                result.actors.append(classes.get_dict_handler(typestr)(e))
        return result

    @property
    def actors(self):
        """
        Obtains the currently set sub-actors.

        :return: the sub-actors
        :rtype: list
        """
        result = self.config["actors"]
        if result is None:
            result = []
        return result

    @actors.setter
    def actors(self, actors):
        """
        Sets the sub-actors of the actor.

        :param actors: the sub-actors
        :type actors: list
        """
        if actors is None:
            actors = self.default_actors()
        self.check_actors(actors)
        self.config["actors"] = actors

    @property
    def active(self):
        """
        Returns the count of non-skipped actors.

        :return: the count
        :rtype: int
        """
        result = 0
        for actor in self.actors:
            if not actor.skip:
                result += 1
        return result

    @property
    def first_active(self):
        """
        Returns the first non-skipped actor.

        :return: the first active actor, None if not available
        :rtype: Actor
        """
        result = None
        for actor in self.actors:
            if not actor.skip:
                result = actor
                break
        return result

    @property
    def last_active(self):
        """
        Returns the last non-skipped actor.

        :return: the last active actor, None if not available
        :rtype: Actor
        """
        result = None
        for actor in reversed(self.actors):
            if not actor.skip:
                result = actor
                break
        return result

    def index_of(self, name):
        """
        Returns the index of the actor with the given name.

        :param name: the name of the Actor to find
        :type name: str
        :return: the index, -1 if not found
        :rtype: int
        """
        result = -1
        for index, actor in enumerate(self.actors):
            if actor.name == name:
                result = index
                break
        return result

    def update_parent(self):
        """
        Updates the parent in its sub-actors.
        """
        for actor in self.actors:
            actor.parent = self

    def setup(self):
        """
        Configures the actor before execution.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        result = super(ActorHandler, self).setup()
        if result is None:
            self.update_parent()
            try:
                self.check_actors(self.actors)
            except Exception as e:
                result = str(e)
        if result is None:
            for actor in self.actors:
                name = actor.name
                newname = actor.unique_name(actor.name)
                if name != newname:
                    actor.name = newname
        if result is None:
            for actor in self.actors:
                if actor.skip:
                    continue
                result = actor.setup()
                if result is not None:
                    break
        if result is None:
            result = self._director.setup()
        return result

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        return self._director.execute()

    def stop_execution(self):
        """
        Triggers the stopping of the actor.
        """
        self._director.stop_execution()

    def wrapup(self):
        """
        Finishes up after execution finishes, does not remove any graphical output.
        """
        for actor in self.actors:
            if actor.skip:
                continue
            actor.wrapup()
        super(ActorHandler, self).wrapup()

    def cleanup(self):
        """
        Destructive finishing up after execution stopped.
        """
        for actor in self.actors:
            if actor.skip:
                continue
            actor.cleanup()
        super(ActorHandler, self).cleanup()

    def _build_tree(self, actor, content):
        """
        Builds the tree for the given actor.

        :param actor: the actor to process
        :type actor: Actor
        :param content: the rows of the tree collected so far
        :type content: list
        """
        depth = actor.depth
        row = ""
        for i in range(depth - 1):
            row += "| "
        if depth > 0:
            row += "|-"
        name = actor.name
        if name != actor.__class__.__name__:
            name = actor.__class__.__name__ + " '" + name + "'"
        row += name
        quickinfo = actor.quickinfo
        if quickinfo is not None:
            row += " [" + quickinfo + "]"
        content.append(row)

        if isinstance(actor, ActorHandler):
            for sub in actor.actors:
                self._build_tree(sub, content)

    @property
    def tree(self):
        """
        Returns a tree representation of this sub-flow.

        :return: the tree
        :rtype: str
        """
        content = []
        self._build_tree(self, content)
        return '\n'.join(content)


class Director(object):
    """
    Ancestor for classes that "direct" the flow of tokens.
    """

    def __init__(self, owner):
        """
        Initializes the director.

        :param owner: the owning actor
        :type owner: Actor
        """
        self.check_owner(owner)
        self._owner = owner

    def __str__(self):
        """
        Returns a short description about the director.

        :return: the description
        :rtype: str
        """
        result = self.__name__
        if self.owner is not None:
            result = result + ", owner=" + self.owner.name
        return result

    @property
    def owner(self):
        """
        Obtains the currently set owner.

        :return: the owner
        :rtype: Actor
        """
        return self._owner

    def check_owner(self, owner):
        """
        Checks the owner. Raises an exception if invalid.

        :param owner: the owner to check
        :type owner: Actor
        """
        pass

    def setup(self):
        """
        Performs some checks.

        :return: None if successful, otherwise error message.
        :rtype: str
        """
        return None

    def do_execute(self):
        """
        Actual execution of the director.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        raise Exception("Not implemented!")

    def execute(self):
        """
        Executes the director.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        if self.owner is None:
            return "No actor set as owner!"
        if self.owner.skip:
            return None
        return self.do_execute()


class SequentialDirector(Director, Stoppable):
    """
    Director for sequential execution of actors.
    """

    def __init__(self, owner):
        """
        Initializes the director.

        :param owner: the owning actor
        :type owner: Actor
        """
        super(SequentialDirector, self).__init__(owner)
        self._stopping = False
        self._stopped = False
        self._allow_source = False
        self._record_output = True
        self._recorded_output = []

    @property
    def allow_source(self):
        """
        Obtains whether to allow a source actor.

        :return: true if to allow
        :rtype: bool
        """
        return self._allow_source

    @allow_source.setter
    def allow_source(self, allow):
        """
        Sets whether to allow a source actor.

        :param allow: true if to allow
        :type allow: bool
        """
        self._allow_source = allow

    @property
    def record_output(self):
        """
        Obtains whether to record the output of the last actor.

        :return: true if to record
        :rtype: bool
        """
        return self._record_output

    @record_output.setter
    def record_output(self, record):
        """
        Sets whether to record the output of the last actor.

        :param record: true if to record
        :type record: bool
        """
        self._record_output = record

    @property
    def recorded_output(self):
        """
        Obtains the recorded output.

        :return: the output
        :rtype: list
        """
        return self._recorded_output

    def stop_execution(self):
        """
        Triggers the stopping of the object.
        """
        if not (self._stopping or self._stopped):
            for actor in self.owner.actors:
                actor.stop_execution()
            self._stopping = True

    def is_stopping(self):
        """
        Returns whether the director is in the process of stopping.

        :return:
        """
        return self._stopping

    def is_stopped(self):
        """
        Returns whether the object has been stopped.

        :return: whether stopped
        :rtype: bool
        """
        return self._stopped

    def check_owner(self, owner):
        """
        Checks the owner. Raises an exception if invalid.

        :param owner: the owner to check
        :type owner: Actor
        """
        if not isinstance(owner, ActorHandler):
            raise Exception("Owner is not an ActorHandler: " + owner.__name__)

    def check_actors(self):
        """
        Checks the actors of the owner. Raises an exception if invalid.
        """
        actors = []
        for actor in self.owner.actors:
            if actor.skip:
                continue
            actors.append(actor)
        if len(actors) == 0:
            return
        if not self.allow_source and base.is_source(actors[0]):
            raise Exception("Actor '" + actors[0].full_name + "' is a source, but no sources allowed!")
        for i in range(1, len(actors)):
            if not isinstance(actors[i], InputConsumer):
                raise Exception("Actor does not accept any input: " + actors[i].full_name)

    def setup(self):
        """
        Performs some checks.

        :return: None if successful, otherwise error message.
        :rtype: str
        """
        result = super(SequentialDirector, self).setup()
        if result is None:
            try:
                self.check_actors()
            except Exception as e:
                result = str(e)
        return result

    def do_execute(self):
        """
        Actual execution of the director.

        :return: None if successful, otherwise error message
        :rtype: str
        """

        self._stopped = False
        self._stopping = False
        not_finished_actor = self.owner.first_active
        pending_actors = []
        finished = False
        actor_result = None

        while not (self.is_stopping() or self.is_stopped()) and not finished:
            # determing starting point of next iteration
            if len(pending_actors) > 0:
                start_index = self.owner.index_of(pending_actors[-1].name)
            else:
                start_index = self.owner.index_of(not_finished_actor.name)
                not_finished_actor = None

            # iterate over actors
            token = None
            last_active = -1
            if self.owner.active > 0:
                last_active = self.owner.last_active.index
            for i in range(start_index, last_active + 1):
                # do we have to stop the execution?
                if self.is_stopped() or self.is_stopping():
                    break

                curr = self.owner.actors[i]
                if curr.skip:
                    continue

                # no token? get pending one or produce new one
                if token is None:
                    if isinstance(curr, OutputProducer) and curr.has_output():
                        pending_actors.pop()
                    else:
                        actor_result = curr.execute()
                        if actor_result is not None:
                            self.owner.logger.error(
                                curr.full_name + " generated following error output:\n" + actor_result)
                            break

                    if isinstance(curr, OutputProducer) and curr.has_output():
                        token = curr.output()
                    else:
                        token = None

                    # still more to come?
                    if isinstance(curr, OutputProducer) and curr.has_output():
                        pending_actors.append(curr)

                else:
                    # process token
                    curr.input = token
                    actor_result = curr.execute()
                    if actor_result is not None:
                        self.owner.logger.error(
                            curr.full_name + " generated following error output:\n" + actor_result)
                        break

                    # was a new token produced?
                    if isinstance(curr, OutputProducer):
                        if curr.has_output():
                            token = curr.output()
                        else:
                            token = None

                        # still more to come?
                        if curr.has_output():
                            pending_actors.append(curr)
                    else:
                        token = None

                # token from last actor generated? -> store
                if (i == self.owner.last_active.index) and (token is not None):
                    if self._record_output:
                        self._recorded_output.append(token)

                # no token produced, ignore rest of actors
                if isinstance(curr, OutputProducer) and (token is None):
                    break

            # all actors finished?
            finished = (not_finished_actor is None) and (len(pending_actors) == 0)

        return actor_result


class Flow(ActorHandler, StorageHandler):
    """
    Root actor for defining and executing flows.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the sequence.

        :param name: the name of the sequence
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(Flow, self).__init__(name=name, config=config)
        self._storage = {}

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Root actor for defining and executing flows."

    def new_director(self):
        """
        Creates the director to use for handling the sub-actors.

        :return: the director instance
        :rtype: Director
        """
        result = SequentialDirector(self)
        result.record_output = False
        result.allow_source = True
        return result

    def check_actors(self, actors):
        """
        Performs checks on the actors that are to be used. Raises an exception if invalid setup.

        :param actors: the actors to check
        :type actors: list
        """
        super(Flow, self).check_actors(actors)
        actor = self.first_active
        if (actor is not None) and not base.is_source(actor):
            raise Exception("First active actor is not a source: " + actor.full_name)

    @property
    def storage(self):
        """
        Returns the internal storage.

        :return: the internal storage
        :rtype: dict
        """
        return self._storage

    @classmethod
    def load(cls, fname):
        """
        Loads the flow from a JSON file.

        :param fname: the file to load
        :type fname: str
        :return: the flow
        :rtype: Flow
        """
        with open(fname) as f:
            content = f.readlines()
        return Flow.from_json(''.join(content))

    @classmethod
    def save(cls, flow, fname):
        """
        Saves the flow to a JSON file.

        :param flow: the flow to save
        :type flow: Flow
        :param fname: the file to load
        :type fname: str
        :return: None if successful, otherwise error message
        :rtype: str
        """
        result = None
        try:
            f = open(fname, 'w')
            f.write(flow.to_json())
            f.close()
        except Exception as e:
            result = str(e)
        return result


class Sequence(ActorHandler, InputConsumer):
    """
    Simple sequence of actors that get executed one after the other. Accepts input.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the sequence.

        :param name: the name of the sequence
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(Sequence, self).__init__(name=name, config=config)
        self._output = []

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Simple sequence of actors that get executed one after the other. Accepts input."

    def new_director(self):
        """
        Creates the director to use for handling the sub-actors.

        :return: the director instance
        :rtype: Director
        """
        result = SequentialDirector(self)
        result.record_output = False
        result.allow_source = False
        return result

    def check_actors(self, actors):
        """
        Performs checks on the actors that are to be used. Raises an exception if invalid setup.

        :param actors: the actors to check
        :type actors: list
        """
        super(Sequence, self).check_actors(actors)
        actor = self.first_active
        if (actor is not None) and not isinstance(actor, InputConsumer):
            raise Exception("First active actor does not accept input: " + actor.full_name)

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        self.first_active.input = self.input
        result = self._director.execute()
        if result is None:
            self._output.append(self.input)
        return result


class Tee(ActorHandler, Transformer):
    """
    'Tees off' the current token to be processed in the sub-tree before passing it on.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the sequence.

        :param name: the name of the sequence
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(Tee, self).__init__(name=name, config=config)
        self._requires_active_actors = True

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "'Tees off' the current token to be processed in the sub-tree before passing it on."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        cond = str(self.config["condition"])
        if len(cond) > 0:
            return "condition: " + cond
        else:
            return None

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(Tee, self).fix_config(options)

        opt = "condition"
        if opt not in options:
            options[opt] = "True"
        if opt not in self.help:
            self.help[opt] = "The (optional) condition for teeing off the tokens; uses the 'eval' method, "\
                             "ie the expression must evaluate to a boolean value; storage values placeholders "\
                             "'@{...}' get replaced with their string representations before evaluating the "\
                             "expression (string)."

        return options

    def new_director(self):
        """
        Creates the director to use for handling the sub-actors.

        :return: the director instance
        :rtype: Director
        """
        result = SequentialDirector(self)
        result.record_output = False
        result.allow_source = False
        return result

    def check_actors(self, actors):
        """
        Performs checks on the actors that are to be used. Raises an exception if invalid setup.

        :param actors: the actors to check
        :type actors: list
        """
        super(Tee, self).check_actors(actors)
        actor = self.first_active
        if actor is None:
            if self._requires_active_actors:
                raise Exception("No active actor!")
        elif not isinstance(actor, InputConsumer):
            raise Exception("First active actor does not accept input: " + actor.full_name)

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        result = None
        teeoff = True
        cond = self.storagehandler.expand(str(self.resolve_option("condition")))
        if len(cond) > 0:
            teeoff = bool(eval(cond))
        if teeoff:
            self.first_active.input = self.input
            result = self._director.execute()
        if result is None:
            self._output.append(self.input)
        return result


class Trigger(ActorHandler, Transformer):
    """
    'Triggers' the sub-tree with each token passing through and then passes the token on.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the sequence.

        :param name: the name of the sequence
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(Trigger, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "'Triggers' the sub-tree with each token passing through and then passes the token on."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        cond = str(self.resolve_option("condition"))
        if len(cond) > 0:
            return "condition: " + cond
        else:
            return None

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(Trigger, self).fix_config(options)

        opt = "condition"
        if opt not in options:
            options[opt] = "True"
        if opt not in self.help:
            self.help[opt] = "The (optional) condition for teeing off the tokens; uses the 'eval' method, "\
                             "ie the expression must evaluate to a boolean value; storage values placeholders "\
                             "'@{...}' get replaced with their string representations before evaluating the "\
                             "expression (string)."

        return options

    def new_director(self):
        """
        Creates the director to use for handling the sub-actors.

        :return: the director instance
        :rtype: Director
        """
        result = SequentialDirector(self)
        result.record_output = False
        result.allow_source = True
        return result

    def check_actors(self, actors):
        """
        Performs checks on the actors that are to be used. Raises an exception if invalid setup.

        :param actors: the actors to check
        :type actors: list
        """
        super(Trigger, self).check_actors(actors)
        actor = self.first_active
        if actor is None:
            raise Exception("No active actor!")
        if not base.is_source(actor):
            raise Exception("First active actor is not a source: " + actor.full_name)

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        result = None
        teeoff = True
        cond = self.storagehandler.expand(str(self.resolve_option("condition")))
        if len(cond) > 0:
            teeoff = bool(eval(cond))
        if teeoff:
            result = self._director.execute()
        if result is None:
            self._output.append(self.input)
        return result


class BranchDirector(Director, Stoppable):
    """
    Director for the Branch actor.
    """

    def __init__(self, owner):
        """
        Initializes the director.

        :param owner: the owning actor
        :type owner: Actor
        """
        super(BranchDirector, self).__init__(owner)
        self._stopping = False
        self._stopped = False

    def stop_execution(self):
        """
        Triggers the stopping of the object.
        """
        if not (self._stopping or self._stopped):
            for actor in self.owner.actors:
                actor.stop_execution()
            self._stopping = True

    def is_stopping(self):
        """
        Returns whether the director is in the process of stopping.

        :return:
        """
        return self._stopping

    def is_stopped(self):
        """
        Returns whether the object has been stopped.

        :return: whether stopped
        :rtype: bool
        """
        return self._stopped

    def check_owner(self, owner):
        """
        Checks the owner. Raises an exception if invalid.

        :param owner: the owner to check
        :type owner: Actor
        """
        if not isinstance(owner, Branch):
            raise Exception("Owner is not a Branch: " + owner.__name__)

    def check_actors(self):
        """
        Checks the actors of the owner. Raises an exception if invalid.
        """
        actors = []
        for actor in self.owner.actors:
            if actor.skip:
                continue
            actors.append(actor)
        if len(actors) == 0:
            return
        for actor in actors:
            if not isinstance(actor, InputConsumer):
                raise Exception("Actor does not accept any input: " + actor.full_name)

    def setup(self):
        """
        Performs some checks.

        :return: None if successful, otherwise error message.
        :rtype: str
        """
        result = super(BranchDirector, self).setup()
        if result is None:
            try:
                self.check_actors()
            except Exception as e:
                result = str(e)
        return result

    def do_execute(self):
        """
        Actual execution of the director.

        :return: None if successful, otherwise error message
        :rtype: str
        """

        result = None
        self._stopped = False
        self._stopping = False

        for actor in self.owner.actors:
            if self.is_stopping() or self.is_stopped():
                break
            actor.input = self.owner.input
            result = actor.execute()
            if result is not None:
                break

        return result


class Branch(ActorHandler, InputConsumer):
    """
    Passes on the input token to all of its sub-actors, one after the other.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the branch.

        :param name: the name of the branch
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(Branch, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Passes on the input token to all of its sub-actors, one after the other."

    def new_director(self):
        """
        Creates the director to use for handling the sub-actors.

        :return: the director instance
        :rtype: Director
        """
        result = BranchDirector(self)
        return result


class ContainerValuePicker(Tee):
    """
    Picks the specified value from the container and 'tees' it off.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the actor.

        :param name: the name of the actor
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(ContainerValuePicker, self).__init__(name=name, config=config)
        self._requires_active_actors = False

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Picks the specified value from the container and 'tees' it off."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        return "value: " + str(self.config["value"]) + ", switch: " + str(self.config["switch"])

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(ContainerValuePicker, self).fix_config(options)

        opt = "value"
        if opt not in options:
            options[opt] = "Model"
        if opt not in self.help:
            self.help[opt] = "The name of the container value to pick from the container (string)."

        opt = "switch"
        if opt not in options:
            options[opt] = False
        if opt not in self.help:
            self.help[opt] = "Whether to switch the ouputs, i.e., forward the container to the sub-flow and the " \
                             + "container value to the following actor instead (bool)."

        return options

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        result = None
        cont = self.input.payload
        name = str(self.resolve_option("value"))
        value = cont.get(name)
        switch = bool(self.resolve_option("switch"))
        if switch:
            if self.first_active is not None:
                self.first_active.input = self.input
                result = self._director.execute()
            if result is None:
                self._output.append(Token(value))
        else:
            if self.first_active is not None:
                self.first_active.input = Token(value)
                result = self._director.execute()
            if result is None:
                self._output.append(self.input)
        return result


class Stop(InputConsumer):
    """
    Stops the execution of the flow.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the branch.

        :param name: the name of the branch
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(Stop, self).__init__(name=name, config=config)
        super(InputConsumer, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Stops the execution of the flow."

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        self.root.stop_execution()
        return None



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# base.py
# Copyright (C) 2015-2016 Fracpete (pythonwekawrapper at gmail dot com)


import logging
import re
import traceback
import uuid
from weka.core.classes import Configurable, Stoppable
import weka.core.classes as classes


class Actor(Configurable, Stoppable):
    """
    The ancestor for all actors.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the actor.

        :param name: the name of the actor
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(Actor, self).__init__(config=config)
        self._name = self.__class__.__name__
        self._parent = None
        self._full_name = None
        self._stopped = False
        if name is not None:
            self.name = name
        if not classes.has_dict_handler("Actor"):
            classes.register_dict_handler("Actor", Actor.from_dict)

    def __str__(self):
        """
        Returns a short representation of the actor's setup.

        :return: the setup
        :rtype: str
        """
        return self.full_name + ": " + str(self._config)

    def __repr__(self):
        """
        Returns Python code for instantiating the object.

        :return: the representation
        :rtype: str
        """
        return \
            self.__class__.__module__ + "." + self.__class__.__name__ \
            + "(name=" + self.name + ", config=" + str(self.config) + ")"

    def new_logger(self):
        """
        Returns a new logger instance.

        :return: the logger instance
        :rtype: logger
        """
        return logging.getLogger(self.full_name)

    @property
    def name(self):
        """
        Obtains the currently set name of the actor.

        :return: the name
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of the actor.

        :param name: the name
        :type name: str
        """
        self._name = name

    def unique_name(self, name):
        """
        Generates a unique name.

        :param name: the name to check
        :type name: str
        :return: the unique name
        :rtype: str
        """
        result = name

        if self.parent is not None:
            index = self.index
            bname = re.sub(r'-[0-9]+$', '', name)
            names = []
            for idx, actor in enumerate(self.parent.actors):
                if idx != index:
                    names.append(actor.name)
            result = bname
            count = 0
            while result in names:
                count += 1
                result = bname + "-" + str(count)

        return result

    @property
    def parent(self):
        """
        Obtains the currently set parent of the actor.

        :return: the name
        :rtype: str
        """
        return self._parent

    @parent.setter
    def parent(self, parent):
        """
        Sets the parent of the actor.

        :param parent: the parent
        :type parent: Actor
        """
        self._name = self.unique_name(self._name)
        self._full_name = None
        self._logger = None
        self._parent = parent

    @property
    def index(self):
        """
        Returns the index of this actor in its parent's list of actors.

        :return: the index, -1 if not available
        :rtype: int
        """
        if self.parent is None:
            return -1
        else:
            return self.parent.index_of(self.name)

    @property
    def full_name(self):
        """
        Obtains the full name of the actor.

        :return: the full name
        :rtype: str
        """
        if self._full_name is None:
            fn = self.name.replace(".", "\\.")
            parent = self._parent
            if parent is not None:
                fn = parent.full_name + "." + fn
            self._full_name = fn

        return self._full_name

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        opt = "annotation"
        if opt not in options:
            options[opt] = None
        if opt not in self.help:
            self.help[opt] = "The (optional) annotation for this actor (string)."

        opt = "skip"
        if opt not in options:
            options[opt] = False
        if opt not in self.help:
            self.help[opt] = "Whether to skip (disable) this actor (bool)."

        return super(Actor, self).fix_config(options)

    def to_dict(self):
        """
        Returns a dictionary that represents this object, to be used for JSONification.

        :return: the object dictionary
        :rtype: dict
        """
        result = super(Actor, self).to_dict()
        result["type"] = "Actor"
        result["name"] = self.name
        return result

    @classmethod
    def from_dict(cls, d):
        """
        Restores an object state from a dictionary, used in de-JSONification.

        :param d: the object dictionary
        :type d: dict
        :return: the object
        :rtype: object
        """
        conf = {}
        for k in d["config"]:
            v = d["config"][k]
            if isinstance(v, dict):
                if u"type" in v:
                    typestr = v[u"type"]
                else:
                    typestr = v["type"]
                conf[str(k)] = classes.get_dict_handler(typestr)(v)
            else:
                conf[str(k)] = v
        return classes.get_class(d["class"])(name=d["name"], config=conf)

    def resolve_option(self, name, default=None):
        """
        Resolves the option, i.e., interprets "@{...}" values and retrievs them instead from internal
        storage.

        :param name: the name of the option
        :type name: str
        :param default: the optional default value
        :type default: object
        :return: the resolved value
        :rtype: object
        """
        value = self.config[name]
        if value is None:
            return default
        elif isinstance(value, str) \
                and value.startswith("@{") \
                and value.endswith("}") \
                and (value.find("@{", 1) == -1):
            stname = value[2:len(value)-1]
            if (self.storagehandler is not None) and (stname in self.storagehandler.storage):
                return self.storagehandler.storage[stname]
            else:
                return default
        else:
            return value

    @property
    def skip(self):
        """
        Obtains whether the actor is disabled (skipped).

        :return: True if skipped
        :rtype: bool
        """
        return self.resolve_option("skip")

    @skip.setter
    def skip(self, skip):
        """
        Sets whether the actor is skipped.

        :param skip: True if skipped
        :type skip: bool
        """
        self.config["skip"] = skip

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        return None

    @property
    def storagehandler(self):
        """
        Returns the storage handler available to thise actor.

        :return: the storage handler, None if not available
        """
        if isinstance(self, StorageHandler):
            return self
        elif self.parent is not None:
            return self.parent.storagehandler
        else:
            return None

    @property
    def root(self):
        """
        Returns the top-level actor.

        :return: the top-level actor
        :rtype: Actor
        """
        if self.parent is None:
            return self
        else:
            return self.parent.root

    @property
    def depth(self):
        """
        Returns the depth of this actor inside the overall flow.

        :return: the depth
        :rtype: int
        """
        if self.parent is None:
            return 0
        else:
            return self.parent.depth + 1

    def is_stopped(self):
        """
        Returns whether the object has been stopped.

        :return: whether stopped
        :rtype: bool
        """
        return self._stopped

    def stop_execution(self):
        """
        Triggers the stopping of the actor.
        """
        self._stopped = True

    def setup(self):
        """
        Configures the actor before execution.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        return None

    def pre_execute(self):
        """
        Gets executed before the actual execution.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        return None

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        raise Exception("Not implemented!")

    def post_execute(self):
        """
        Gets executed after the actual execution.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        return None

    def execute(self):
        """
        Executes the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        if self.skip:
            return None

        result = self.pre_execute()
        if result is None:
            try:
                result = self.do_execute()
            except Exception as e:
                result = traceback.format_exc()
                print(self.full_name + "\n" + result)
        if result is None:
            result = self.post_execute()
        return result

    def wrapup(self):
        """
        Finishes up after execution finishes, does not remove any graphical output.
        """
        pass

    def cleanup(self):
        """
        Destructive finishing up after execution stopped.
        """
        pass


class Token(object):
    """
    Container for transporting data through the flow.
    """

    def __init__(self, payload):
        """
        Initializes the token with the given payload.

        :param payload: the payload for the token.
        :type payload: object
        """
        self._id = str(uuid.uuid4())
        self._payload = payload

    @property
    def id(self):
        """
        Obtains the ID of the token.

        :return: the ID
        :rtype: str
        """
        return self._id

    @property
    def payload(self):
        """
        Obtains the currently set payload.

        :return: the payload
        :rtype: object
        """
        return self._payload

    def __str__(self):
        """
        Returns a short representation of the token and its payload.
        """
        return self._id + ": " + str(self._payload)


class InputConsumer(Actor):
    """
    Actors that consume tokens inherit this class.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the actor.

        :param name: the name of the actor
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(InputConsumer, self).__init__(name=name, config=config)
        self._input = None

    def check_input(self, token):
        """
        Performs checks on the input token. Raises an exception if unsupported.

        :param token: the token to check
        :type token: Token
        """
        pass

    @property
    def input(self):
        """
        Returns the current input token, None if not available.

        :return: the input token
        :rtype: Token
        """
        return self._input

    @input.setter
    def input(self, token):
        """
        Accepts the data for processing.

        :param token: the token to process
        :type token: Token
        """
        self.check_input(token)
        self._input = token


class OutputProducer(Actor):
    """
    Actors that generate output tokens inherit this class.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the actor.

        :param name: the name of the actor
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(OutputProducer, self).__init__(name=name, config=config)
        self._output = None

    def pre_execute(self):
        """
        Gets executed before the actual execution.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        self._output = []
        return None

    def has_output(self):
        """
        Checks whether any output tokens are present.

        :return: true if at least one output token present
        :rtype: bool
        """
        return (self._output is not None) and (len(self._output) > 0)

    def output(self):
        """
        Returns the next available output token.

        :return: the next token, None if none available
        :rtype: Token
        """
        if (self._output is None) or (len(self._output) == 0):
            result = None
        else:
            result = self._output.pop(0)
        return result


class StorageHandler(object):
    """
    For classes that support internal storage (= dictionary).
    """

    @property
    def storage(self):
        """
        Returns the internal storage.

        :return: the internal storage
        :rtype: dict
        """
        raise Exception("Not implemented!")

    def expand(self, s):
        """
        Expands all occurrences of "@{...}" within the string with the actual values currently stored
        in internal storage.

        :param s: the string to expand
        :type s: str
        :return: the expanded string
        :rtype: str
        """
        result = s
        while result.find("@{") > -1:
            start = result.index("@{")
            end = result.index("}", start)
            name = result[start + 2:end]
            value = self.storage[name]
            if value is None:
                raise("Storage value '" + name + "' not present, failed to expand string: " + s)
            else:
                result = result[0:start] + str(value) + result[end + 1:]
        return result

    @classmethod
    def pad(cls, name):
        """
        Pads the name with "@{...}".

        :param name: the name to pad
        :type name: str
        :return: the padded name
        :rtype: str
        """
        if name.startswith("@{"):
            return name
        else:
            return "@{" + name + "}"

    @classmethod
    def extract(cls, padded):
        """
        Removes the surrounding "@{...}" from the name.

        :param padded: the padded string
        :type padded: str
        :return: the extracted name
        :rtype: str
        """
        if padded.startswith("@{") and padded.endswith("}"):
            return padded[2:len(padded)-1]
        else:
            return padded


def is_source(actor):
    """
    Checks whether the actor is a source.

    :param actor: the actor to check
    :type actor: Actor
    :return: True if the actor is a source
    :rtype: bool
    """
    return not isinstance(actor, InputConsumer) and isinstance(actor, OutputProducer)


def is_transformer(actor):
    """
    Checks whether the actor is a transformer.

    :param actor: the actor to check
    :type actor: Actor
    :return: True if the actor is a transformer
    :rtype: bool
    """
    return isinstance(actor, InputConsumer) and isinstance(actor, OutputProducer)


def is_sink(actor):
    """
    Checks whether the actor is a sink.

    :param actor: the actor to check
    :type actor: Actor
    :return: True if the actor is a sink
    :rtype: bool
    """
    return isinstance(actor, InputConsumer) and not isinstance(actor, OutputProducer)


def to_commandline(o):
    """
    Turns the object into a commandline string. However, first checks whether a string represents
    a internal value placeholder (@{...}).

    :param o: the object to turn into commandline
    :type o: object
    :return: the commandline
    :rtype: str
    """
    if isinstance(o, str) and o.startswith("@{") and o.endswith("}"):
        return o
    else:
        return classes.to_commandline(o)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# weka_test.py
# Copyright (C) 2014-2016 Fracpete (pythonwekawrapper at gmail dot com)

import unittest
import tempfile
import os


class WekaTest(unittest.TestCase):
    """
    Weka-specific test case class.
    """

    def tempdir(self):
        """
        Returns the temp directory.
        :return: the temp directory
        :rtype: str
        """
        return tempfile.gettempdir()

    def tempfile(self, fname):
        """
        Generates a full path for the given filename (without path) and returns it.
        :param fname: the filename (without path)
        :type fname: str
        :return: the full path
        :rtype: str
        """
        return self.tempdir() + os.sep + fname

    def delfile(self, path):
        """
        Removes the file from the file system.
        :param path: the full path to the file
        :type path: str
        :return: whether successfully removed (also True if not present)
        :rtype: bool
        """
        if os.path.exists(path):
            try:
                os.remove(path)
                return True
            except Exception as e:
                print("Failed to remove: " + path + "\n" + str(e))
                return False
        else:
            return True

    def datadir(self):
        """
        Returns the data directory.
        :return: the data directory
        :rtype: str
        """
        rootdir = os.path.dirname(__file__)
        libdir = rootdir + os.sep + ".." + os.sep + "data"
        return libdir

    def datafile(self, fname):
        """
        Returns the full path for the file (without path) and returns it.
        :param fname: the filename (without path)
        :type fname: str
        :return: the full path
        :rtype: str
        """
        return self.datadir() + os.sep + fname



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# experiments.py
# Copyright (C) 2014-2018 Fracpete (pythonwekawrapper at gmail dot com)

import unittest
import weka.core.jvm as jvm
import weka.core.converters as converters
import weka.experiments as experiments
import weka.classifiers as classifiers
import wekatests.tests.weka_test as weka_test


class TestExperiments(weka_test.WekaTest):

    def test_crossvalidation_classification(self):
        """
        Tests cross-validated classification.
        """
        datasets = [self.datafile("iris.arff"), self.datafile("anneal.arff")]
        cls = [
            classifiers.Classifier(classname="weka.classifiers.rules.ZeroR"),
            classifiers.Classifier(classname="weka.classifiers.trees.J48")]
        outfile = self.tempfile("results-cv.arff")
        exp = experiments.SimpleCrossValidationExperiment(
            classification=True,
            runs=10,
            folds=10,
            datasets=datasets,
            classifiers=cls,
            result=outfile)
        self.assertIsNotNone(exp, msg="Failed to instantiate!")
        exp.setup()
        exp.run()

        # evaluate
        loader = converters.loader_for_file(outfile)
        data = loader.load_file(outfile)
        self.assertIsNotNone(data, msg="Failed to load data: " + outfile)

        matrix = experiments.ResultMatrix(classname="weka.experiment.ResultMatrixPlainText")
        self.assertIsNotNone(matrix, msg="Failed to instantiate!")

        tester = experiments.Tester(classname="weka.experiment.PairedCorrectedTTester")
        self.assertIsNotNone(tester, msg="Failed to instantiate!")

        tester.resultmatrix = matrix
        comparison_col = data.attribute_by_name("Percent_correct").index
        tester.instances = data
        self.assertGreater(len(tester.header(comparison_col)), 0, msg="Generated no header")
        self.assertGreater(len(tester.multi_resultset_full(0, comparison_col)), 0, msg="Generated no result")

    def test_randomsplit_regression(self):
        """
        Tests random split on regression.
        """
        datasets = [self.datafile("bolts.arff"), self.datafile("bodyfat.arff")]
        cls = [
            classifiers.Classifier(classname="weka.classifiers.rules.ZeroR"),
            classifiers.Classifier(classname="weka.classifiers.functions.LinearRegression")
        ]
        outfile = self.tempfile("results-rs.arff")
        exp = experiments.SimpleRandomSplitExperiment(
            classification=False,
            runs=10,
            percentage=66.6,
            preserve_order=False,
            datasets=datasets,
            classifiers=cls,
            result=outfile)
        self.assertIsNotNone(exp, msg="Failed to instantiate!")
        exp.setup()
        exp.run()

        # evaluate
        loader = converters.loader_for_file(outfile)
        data = loader.load_file(outfile)
        self.assertIsNotNone(data, msg="Failed to load data: " + outfile)

        matrix = experiments.ResultMatrix(classname="weka.experiment.ResultMatrixPlainText")
        self.assertIsNotNone(matrix, msg="Failed to instantiate!")

        tester = experiments.Tester(classname="weka.experiment.PairedCorrectedTTester")
        self.assertIsNotNone(tester, msg="Failed to instantiate!")

        tester.resultmatrix = matrix
        comparison_col = data.attribute_by_name("Correlation_coefficient").index
        tester.instances = data
        self.assertGreater(len(tester.header(comparison_col)), 0, msg="Generated no header")
        self.assertGreater(len(tester.multi_resultset_full(0, comparison_col)), 0, msg="Generated no result")

    def test_result_matrix(self):
        """
        Tests the ResultMatrix class.
        """
        datasets = [self.datafile("iris.arff"), self.datafile("anneal.arff")]
        cls = [
            classifiers.Classifier(classname="weka.classifiers.rules.ZeroR"),
            classifiers.Classifier(classname="weka.classifiers.trees.J48")]
        outfile = self.tempfile("results-cv.arff")
        exp = experiments.SimpleCrossValidationExperiment(
            classification=True,
            runs=10,
            folds=10,
            datasets=datasets,
            classifiers=cls,
            result=outfile)
        self.assertIsNotNone(exp, msg="Failed to instantiate!")
        exp.setup()
        exp.run()

        # evaluate
        loader = converters.loader_for_file(outfile)
        data = loader.load_file(outfile)
        self.assertIsNotNone(data, msg="Failed to load data: " + outfile)

        matrix = experiments.ResultMatrix(classname="weka.experiment.ResultMatrixPlainText")
        self.assertIsNotNone(matrix, msg="Failed to instantiate!")

        tester = experiments.Tester(classname="weka.experiment.PairedCorrectedTTester")
        self.assertIsNotNone(tester, msg="Failed to instantiate!")

        tester.resultmatrix = matrix
        comparison_col = data.attribute_by_name("Percent_correct").index
        tester.instances = data
        self.assertGreater(len(tester.header(comparison_col)), 0, msg="Generated no header")
        self.assertGreater(len(tester.multi_resultset_full(0, comparison_col)), 0, msg="Generated no result")

        # dimensions
        self.assertEqual(2, matrix.rows, msg="# of rows differ")
        self.assertEqual(2, matrix.columns, msg="# of rows differ")

        # cols
        #self.assertTrue(matrix.get_col_name(0).find("ZeroR") > -1, msg="ZeroR should be part of col name")
        #self.assertTrue(matrix.get_col_name(1).find("J48") > -1, msg="J48 should be part of col name")
        #self.assertIsNone(matrix.get_col_name(2), msg="Column name should not exist")
        #matrix.set_col_name(0, "zeror")
        #self.assertTrue(matrix.get_col_name(0).find("zeror") > -1, msg="zeror should be part of col name")

        self.assertFalse(matrix.is_col_hidden(1), msg="Column should be visible")
        matrix.hide_col(1)
        self.assertTrue(matrix.is_col_hidden(1), msg="Column should be hidden")
        matrix.show_col(1)
        self.assertFalse(matrix.is_col_hidden(1), msg="Column should be visible again")

        # rows
        #self.assertEqual("iris", matrix.get_row_name(0), msg="Row names differ")
        #self.assertEqual("anneal", matrix.get_row_name(1), msg="Row names differ")
        #self.assertIsNone(matrix.get_col_name(2), msg="Row name should not exist")
        #matrix.set_row_name(0, "IRIS")
        #self.assertEqual("IRIS", matrix.get_row_name(0), msg="Row names differ")

        self.assertFalse(matrix.is_row_hidden(1), msg="Row should be visible")
        matrix.hide_row(1)
        self.assertTrue(matrix.is_row_hidden(1), msg="Row should be hidden")
        matrix.show_row(1)
        self.assertFalse(matrix.is_row_hidden(1), msg="Row should be visible again")

        # mean
        self.assertAlmostEqual(33.3, matrix.get_mean(0, 0), places=1, msg="Means differ")
        self.assertAlmostEqual(54.75, matrix.average(0), places=2, msg="Averages differ")
        matrix.set_mean(0, 0, 10)
        self.assertAlmostEqual(10.0, matrix.get_mean(0, 0), places=1, msg="Means differ")

        # stdev
        self.assertAlmostEqual(0.0, matrix.get_stdev(0, 0), places=1, msg="Means differ")
        matrix.set_stdev(0, 0, 0.3)
        self.assertAlmostEqual(0.3, matrix.get_stdev(0, 0), places=1, msg="Means differ")


def suite():
    """
    Returns the test suite.
    :return: the test suite
    :rtype: unittest.TestSuite
    """
    return unittest.TestLoader().loadTestsFromTestCase(TestExperiments)


if __name__ == '__main__':
    jvm.start()
    unittest.TextTestRunner().run(suite())
    jvm.stop()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# stopwords.py
# Copyright (C) 2015-2016 Fracpete (pythonwekawrapper at gmail dot com)

import javabridge
from weka.core.classes import OptionHandler


class Stopwords(OptionHandler):
    """
    Wrapper class for stopwords handlers.
    """

    def __init__(self, classname="weka.core.stopwords.Null", jobject=None, options=None):
        """
        Initializes the specified stopwords handler using either the classname or the supplied JB_Object.

        :param classname: the classname of the stopwords handler
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        if jobject is None:
            jobject = Stopwords.new_instance(classname)
        self.enforce_type(jobject, "weka.core.stopwords.StopwordsHandler")
        super(Stopwords, self).__init__(jobject=jobject, options=options)
        self.__is_stopword = javabridge.make_call(self.jobject, "isStopword", "(Ljava/lang/String;)Z")

    def is_stopword(self, s):
        """
        Checks a string whether it is a stopword.

        :param s: the string to check
        :type s: str
        :return: True if a stopword
        :rtype: bool
        """
        return self.__is_stopword(javabridge.get_env().new_string_utf(s))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# setup.py
# Copyright (C) 2014-2021 Fracpete (pythonwekawrapper at gmail dot com)

from setuptools import setup


def _read(f):
    """
    Reads in the content of the file.
    :param f: the file to read
    :type f: str
    :return: the content
    :rtype: str
    """
    return open(f, 'rb').read()


setup(
    name="python-weka-wrapper3",
    description="Python3 wrapper for the Weka Machine Learning Workbench",
    long_description=(
        _read('DESCRIPTION.rst') + b'\n' +
        _read('CHANGES.rst')).decode('utf-8'),
    url="https://github.com/fracpete/python-weka-wrapper3",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Programming Language :: Python :: 3',
    ],
    license='GNU General Public License version 3.0 (GPLv3)',
    package_dir={
        '': 'python'
    },
    packages=[
        "weka",
        "weka.core",
        "weka.flow",
        "weka.plot"
    ],
    package_data={
        "weka": ["lib/*.jar"],
    },
    include_package_data=True,
    version="0.2.0",
    author='Peter "fracpete" Reutemann',
    author_email='pythonwekawrapper@gmail.com',
    install_requires=[
        "javabridge>=1.0.14",
        "numpy",
    ],
    extras_require={
        'plots': ["matplotlib"],
        'graphs': ["pygraphviz", "PIL"],
    },
    entry_points={
        "console_scripts": [
            "pww-associator=weka.associations:sys_main",
            "pww-attsel=weka.attribute_selection:sys_main",
            "pww-classifier=weka.classifiers:sys_main",
            "pww-clusterer=weka.clusterers:sys_main",
            "pww-datagenerator=weka.datagenerators:sys_main",
            "pww-filter=weka.filters:sys_main",
        ]
    }
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# dataset.py
# Copyright (C) 2014-2019 Fracpete (pythonwekawrapper at gmail dot com)

import unittest
import weka.core.jvm as jvm
import weka.core.dataset as dataset
import weka.core.converters as converters
import wekatests.tests.weka_test as weka_test
from weka.core.dataset import create_instances_from_lists, create_instances_from_matrices
from random import randint
import numpy as np


class TestDataset(weka_test.WekaTest):

    def test_attribute(self):
        """
        Tests the Attribute class.
        """
        name = "Num"
        att = dataset.Attribute.create_numeric(name)
        self.assertIsNotNone(att, "Failed to create attribute!")
        self.assertEqual(name, att.name, "Names differ")
        self.assertTrue(att.is_numeric)

        name = "Nom"
        att = dataset.Attribute.create_nominal(name, ["A", "B", "C"])
        self.assertIsNotNone(att, "Failed to create attribute!")
        self.assertEqual(name, att.name, "Names differ")
        self.assertTrue(att.is_nominal)

        name = "Dat1"
        att = dataset.Attribute.create_date(name)
        self.assertIsNotNone(att, "Failed to create attribute!")
        self.assertEqual(name, att.name, "Names differ")
        self.assertTrue(att.is_date)

        name = "Dat2"
        att = dataset.Attribute.create_date(name, formt="yyyy-MM-dd HH:mm")
        self.assertIsNotNone(att, "Failed to create attribute!")
        self.assertEqual(name, att.name, "Names differ")
        self.assertTrue(att.is_date)

        name = "Str"
        att = dataset.Attribute.create_string(name)
        self.assertIsNotNone(att, "Failed to create attribute!")
        self.assertEqual(name, att.name, "Names differ")
        self.assertTrue(att.is_string)

    def test_attributestats(self):
        """
        Tests the AttributeStats class.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data, msg="Failed to load data!")

        stats = data.attribute_stats(2)
        self.assertIsNotNone(stats, msg="Failed to obtain stats!")
        self.assertEqual(8, stats.distinct_count, "distinct_count differs")
        self.assertEqual(898, stats.int_count, "int_count differs")
        self.assertEqual(0, stats.missing_count, "missing_count differs")
        self.assertEqual([86, 256, 440, 0, 51, 20, 10, 19, 16], stats.nominal_counts.tolist(), "nominal_counts differs")
        self.assertEqual([86, 256, 440, 0, 51, 20, 10, 19, 16], stats.nominal_weights.tolist(), "nominal_weights differs")
        self.assertEqual(898, stats.total_count, "total_count differs")
        self.assertEqual(0, stats.unique_count, "unique_count differs")

    def test_stats(self):
        """
        Tests the Stats class.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data, msg="Failed to load data!")

        stats = data.attribute_stats(3)
        numstats = stats.numeric_stats
        self.assertAlmostEqual(898, numstats.count, places=3, msg="count differs")
        self.assertAlmostEqual(70, numstats.max, places=3, msg="max differs")
        self.assertAlmostEqual(3.635, numstats.mean, places=3, msg="mean differs")
        self.assertAlmostEqual(0.0, numstats.min, places=3, msg="min differs")
        self.assertAlmostEqual(13.717, numstats.stddev, places=3, msg="stddev differs")
        self.assertAlmostEqual(3264, numstats.sum, places=3, msg="sum differs")
        self.assertAlmostEqual(180636, numstats.sumsq, places=3, msg="sumsq differs")

    def test_instance(self):
        """
        Tests the Instance class.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.ORIG.arff"))
        self.assertIsNotNone(data, msg="Failed to load data!")

        inst = data.get_instance(0)
        self.assertEqual(39, inst.num_attributes, msg="num_attributes differs")
        self.assertEqual(-1, data.class_index, msg="class_index differs")

        data.class_index = data.num_attributes - 1
        self.assertEqual(38, data.class_index, msg="class_index differs")

        data.class_is_first()
        self.assertEqual(0, data.class_index, msg="class_index differs")

        data.class_is_last()
        self.assertEqual(38, data.class_index, msg="class_index differs")

        self.assertIsNotNone(inst.dataset, msg="Dataset reference should not be None!")
        self.assertTrue(inst.has_missing(), msg="Should have missing values")
        self.assertTrue(inst.is_missing(0), msg="First value should be missing")
        self.assertFalse(inst.is_missing(1), msg="Second value should not be missing")

        self.assertEqual("C", inst.get_string_value(1), msg="string value differs")
        inst.set_string_value(1, "H")
        self.assertEqual("H", inst.get_string_value(1), msg="string value differs")

        self.assertEqual(8, inst.get_value(3), msg="numeric value differs")
        inst.set_value(3, 6.3)
        self.assertEqual(6.3, inst.get_value(3), msg="numeric value differs")

        self.assertEqual(1, inst.weight, msg="weight should be 1")
        inst.weight = 0.5
        self.assertEqual(0.5, inst.weight, msg="weights differ")

        values = [1.0, 2.0, 3.0]
        inst = dataset.Instance.create_instance(values, classname="weka.core.DenseInstance")
        self.assertEqual(3, inst.num_attributes, msg="#attributes differ")
        self.assertEqual(1.0, inst.get_value(0), msg="value at #" + str(0) + " differs")
        self.assertEqual(2.0, inst.get_value(1), msg="value at #" + str(1) + " differs")
        self.assertEqual(3.0, inst.get_value(2), msg="value at #" + str(2) + " differs")

        values = [0.0, 2.0, 0.0]
        inst = dataset.Instance.create_instance(values, classname="weka.core.SparseInstance")
        self.assertEqual(3, inst.num_attributes, msg="#attributes differ")
        self.assertEqual(0.0, inst.get_value(0), msg="value at #" + str(0) + " differs")
        self.assertEqual(2.0, inst.get_value(1), msg="value at #" + str(1) + " differs")
        self.assertEqual(0.0, inst.get_value(2), msg="value at #" + str(2) + " differs")

        values = [(1, 2.0)]
        inst = dataset.Instance.create_sparse_instance(values, 3, classname="weka.core.SparseInstance")
        self.assertEqual(3, inst.num_attributes, msg="#attributes differ")
        self.assertEqual(0.0, inst.get_value(0), msg="value at #" + str(0) + " differs")
        self.assertEqual(2.0, inst.get_value(1), msg="value at #" + str(1) + " differs")
        self.assertEqual(0.0, inst.get_value(2), msg="value at #" + str(2) + " differs")

    def test_instances(self):
        """
        Tests the Instances class.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data, msg="Failed to load data!")
        self.assertFalse(data.has_class(), msg="Should not have class set!")

        count = 0
        for i in data:
            count += 1
        self.assertEqual(898, count, msg="Number of rows differs!")

        count = 0
        for i in data.attributes():
            count += 1
        self.assertEqual(39, count, msg="Number of attributes differs!")

        self.assertEqual(898, data.num_instances, msg="num_instances differs")
        self.assertEqual(39, data.num_attributes, msg="num_attributes differs")
        self.assertEqual(-1, data.class_index, msg="class_index differs")

        data.class_index = data.num_attributes - 1
        self.assertEqual(38, data.class_index, msg="class_index differs")
        self.assertEqual(38, data.class_attribute.index, msg="class_index differs")

        data.class_is_first()
        self.assertEqual(0, data.class_index, msg="class_index differs")
        self.assertTrue(data.has_class(), msg="Should have class set!")

        data.class_is_last()
        self.assertEqual(38, data.class_index, msg="class_index differs")
        self.assertTrue(data.has_class(), msg="Should have class set!")

        att = data.attribute(0)
        self.assertIsNotNone(att, msg="Attribute should not be None!")
        self.assertEqual("family", att.name, msg="attribute name differs")

        name = "steel"
        att = data.attribute_by_name(name)
        self.assertIsNotNone(att, msg="Attribute should not be None!")
        self.assertEqual(name, att.name, msg="attribute name differs")
        self.assertEqual(2, att.index, msg="attribute index differs")

        data.delete_attribute(2)
        self.assertEqual(38, data.num_attributes, msg="num_attributes differs")
        name = "steel"
        att = data.attribute_by_name(name)
        self.assertIsNone(att, msg="Attribute should be None!")

        data.delete(3)
        self.assertEqual(897, data.num_instances, msg="num_instances differs")

        data.class_is_last()
        data.no_class()
        self.assertFalse(data.has_class(), msg="Should not have class set!")

        # changing rows
        data1 = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data1, msg="Failed to load data!")
        data2 = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data2, msg="Failed to load data!")

        inst1 = data1.get_instance(0)
        inst2 = data2.get_instance(0)
        self.assertEqual(inst1.values.tolist(), inst2.values.tolist(), msg="values differ")

        data1.add_instance(inst2)
        self.assertEqual(899, data1.num_instances, msg="num_instances differs (add)")

        inst2 = data2.get_instance(2)
        data1.add_instance(inst2, index=10)
        inst1 = data1.get_instance(10)
        self.assertEqual(900, data1.num_instances, msg="num_instances differs (insert)")
        self.assertEqual(inst1.values.tolist(), inst2.values.tolist(), msg="values differ (insert)")

        inst2 = data2.get_instance(1)
        data1.set_instance(0, inst2)
        inst1 = data1.get_instance(0)
        self.assertEqual(inst1.values.tolist(), inst2.values.tolist(), msg="values differ (set)")

        # create instances
        atts = []
        atts.append(dataset.Attribute.create_numeric("num"))
        atts.append(dataset.Attribute.create_nominal("nom", ["yes", "no"]))
        atts.append(dataset.Attribute.create_string("str"))
        atts.append(dataset.Attribute.create_date("dat", "yyyy-MM-dd"))
        data = dataset.Instances.create_instances("created", atts, 3)
        # 1. row
        values = []
        values.append(1.1)
        values.append(data.attribute(1).index_of("no"))
        values.append(data.attribute(2).add_string_value("blah de blah"))
        values.append(data.attribute(3).parse_date("2015-10-12"))
        inst = dataset.Instance.create_instance(values)
        data.add_instance(inst)
        # 2. row
        values = []
        values.append(1.2)
        values.append(data.attribute(1).index_of("yes"))
        values.append(data.attribute(2).add_string_value("pi day!"))
        values.append(data.attribute(3).parse_date("2001-12-31"))
        inst = dataset.Instance.create_instance(values)
        data.add_instance(inst)
        # 3. row
        values = []
        values.append(1.3)
        values.append(data.attribute(1).index_of("no"))
        values.append(data.attribute(2).add_string_value("hello world"))
        values.append(data.attribute(3).parse_date("2011-02-03"))
        inst = dataset.Instance.create_instance(values)
        data.add_instance(inst)

        self.assertEqual(data.get_instance(0).get_value(0), 1.1, msg="Numeric value differs")
        self.assertEqual(data.get_instance(1).get_value(0), 1.2, msg="Numeric value differs")
        self.assertEqual(data.get_instance(2).get_value(0), 1.3, msg="Numeric value differs")
        self.assertEqual(data.get_instance(0).get_string_value(1), "no", msg="Nominal value differs")
        self.assertEqual(data.get_instance(1).get_string_value(1), "yes", msg="Nominal value differs")
        self.assertEqual(data.get_instance(2).get_string_value(1), "no", msg="Nominal value differs")
        self.assertEqual(data.get_instance(0).get_string_value(2), "blah de blah", msg="String value differs")
        self.assertEqual(data.get_instance(1).get_string_value(2), "pi day!", msg="String value differs")
        self.assertEqual(data.get_instance(2).get_string_value(2), "hello world", msg="String value differs")
        self.assertEqual(data.get_instance(0).get_value(3), data.attribute(3).parse_date("2015-10-12"), msg="Date value differs")
        self.assertEqual(data.get_instance(1).get_value(3), data.attribute(3).parse_date("2001-12-31"), msg="Date value differs")
        self.assertEqual(data.get_instance(2).get_value(3), data.attribute(3).parse_date("2011-02-03"), msg="Date value differs")

        # train/test split
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("iris.arff"))
        self.assertIsNotNone(data, msg="Failed to load data!")
        self.assertFalse(data.has_class(), msg="Should not have class set!")
        data.class_is_last()

        perc = 66.6
        train, test = data.train_test_split(perc)
        self.assertIsNotNone(train, msg="Train is None")
        self.assertIsNotNone(test, msg="Test is None")
        self.assertEqual(data.num_instances, train.num_instances + test.num_instances, msg="Total number of instances differ")
        self.assertEqual(train.num_instances, 100, msg="Number of training instances differ")
        self.assertEqual(test.num_instances, 50, msg="Number of test instances differ")
        try:
            perc = 0.0
            data.train_test_split(perc)
            self.fail(msg="Should not accept split percentage of " + str(perc))
        except Exception as e:
            pass
        try:
            perc = -1.0
            data.train_test_split(perc)
            self.fail(msg="Should not accept split percentage of " + str(perc))
        except Exception as e:
            pass
        try:
            perc = 100.0
            data.train_test_split(perc)
            self.fail(msg="Should not accept split percentage of " + str(perc))
        except Exception as e:
            pass
        try:
            perc = 101.0
            data.train_test_split(perc)
            self.fail(msg="Should not accept split percentage of " + str(perc))
        except Exception as e:
            pass

    def test_create_instances_from_lists(self):
        """
        Tests the create_instances_from_lists method.
        """

        # numeric
        x = [[randint(1, 10) for _ in range(5)] for _ in range(10)]
        y = [randint(0, 1) for _ in range(10)]
        self.assertEqual(len(x), 10)
        self.assertEqual(len(y), 10)
        dataset = create_instances_from_lists(x, name="generated from lists (no y)")
        self.assertEqual(len(dataset), 10)
        dataset = create_instances_from_lists(x, y, name="generated from lists")
        self.assertEqual(len(dataset), 10)

        # mixed
        x = [["TEXT", 1, 1.1], ["XXX", 2, 2.2]]
        y = ["A", "B"]
        self.assertEqual(len(x), 2)
        self.assertEqual(len(y), 2)
        dataset = create_instances_from_lists(x, name="generated from mixed lists (no y)")
        self.assertEqual(len(dataset), 2)
        dataset = create_instances_from_lists(x, y, name="generated from mixed lists")
        self.assertEqual(len(dataset), 2)

    def test_create_instances_from_matrices(self):
        """
        Tests the create_instances_from_matrices method.
        """

        # numeric
        x = np.random.randn(10, 5)
        y = np.random.randn(10)
        self.assertEqual(len(x), 10)
        self.assertEqual(len(y), 10)
        dataset = create_instances_from_lists(x, name="generated from lists (no y)")
        self.assertEqual(len(dataset), 10)
        dataset = create_instances_from_lists(x, y, name="generated from lists")
        self.assertEqual(len(dataset), 10)

        # mixed
        x = np.array([("TEXT", 1, 1.1), ("XXX", 2, 2.2)], dtype='S20, i4, f8')
        y = np.array(["A", "B"], dtype='S20')
        self.assertEqual(len(x), 2)
        self.assertEqual(len(y), 2)
        dataset = create_instances_from_matrices(x, name="generated from mixed lists (no y)")
        self.assertEqual(len(dataset), 2)
        dataset = create_instances_from_matrices(x, y, name="generated from mixed lists")
        self.assertEqual(len(dataset), 2)


def suite():
    """
    Returns the test suite.
    :return: the test suite
    :rtype: unittest.TestSuite
    """
    return unittest.TestLoader().loadTestsFromTestCase(TestDataset)


if __name__ == '__main__':
    jvm.start()
    unittest.TextTestRunner().run(suite())
    jvm.stop()
